import os
import shutil
import yaml

from loguru import logger

from .tests import *
from .read_shipped_data import *

# Bold strings
bs = "\033[1m"
be = "\033[0m"


def user_config(info):
    """
    Reads the ``user_config.yaml`` file that contains both the account information and
    the testing directory for the tests. If the file does not exists it asks the user
    for the information.

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info, including the user defined options.

    Returns
    -------
    info : dict
        Same as input but with a nested ``user`` dictionary.
    """
    # If user info is not needed return None
    if info["ignore_user_info"]:
        info["user"] = None
        return info

    # Check for user configuration file
    user_config = f"{info['script_dir']}/user_config.yaml"
    logger.info("")
    if not os.path.isfile(user_config):
        # Make the user configuration file
        answers = {}
        logger.info(
            f"{bs}Welcome to ESM-Tests! Automatic testing for ESM-Tools devs\n"
            + f"**********************************************************{be}\n"
            + "Please answer the following questions. If you ever need to change the "
            + "configuration, you can do that in the the esm_tests/user_config.yaml\n"
        )
        try:
            answers["account"] = input(
                "What account will you be using for testing? (default: None) "
            )
        except EOFError:
            if os.environ.get("CI"):
                logger.info(
                    "This is probably running on the CI System. We will default to None"
                )
                answers["account"] = None
            else:
                raise
        if not answers["account"] or answers["account"] == "None":
            answers["account"] = None
        try:
            answers["test_dir"] = input(
                "In which directory would you like to run the tests? "
            )
        except EOFError:
            if os.environ.get("CI"):
                logger.info(
                    f"This is probably running on the CI System. We will default to {os.getcwd()}"
                )
                answers["test_dir"] = os.getcwd()
            else:
                raise
        with open(user_config, "w") as uc:
            logger.debug("Writing file")
            out = yaml.dump(answers)
            uc.write(out)
            logger.debug(f"Done: {uc}")

    # Load the user info
    with open(user_config, "r") as uc:
        user_info = yaml.load(uc, Loader=yaml.FullLoader)
    logger.info(f"{bs}Running tests with the following configuration:{be}")
    logger.info(f"{bs}-----------------------------------------------{be}")
    yprint(user_info)

    info["user"] = user_info

    return info


def get_scripts(info):
    """
    Get path and names of the runscripts to be tested and store it in the ``info``
    dictionary. If the file ``esm_tools/src/esm_tests/test_config.yaml`` exists, it
    only loads the information from runscripts specified there. Otherwise, it loads all
    the runscripts available in ``esm_tools/src/resources/runscripts``.

    The structur of the ``test_config.yaml`` should be::

    .. code-block:: yaml

        <model1>:       # A list to specify runscript by runscript
            - <runscript1>.yaml
            - <runscript2>.yaml
        <model2>: all   # To run all runscripts of the model
        [ ... ]

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info.

    Returns
    -------
    info : dict
        Same as input but with a nested ``scripts`` dictionary.
    """
    for key, value in info.items():
        logger.debug(f"key {key}: value {value}")
    # Get the path for the testing runscripts
    runscripts_dir = get_runscripts_dir()
    # Initialize the scripts_info dictionary
    scripts_info = {}
    # Counter for calculating percentages
    ns = 0

    # Load test info
    test_config = f"{info.get('script_dir', os.getcwd())}/test_config.yaml"
    # Check if there is a user-defined dictionary, where the user has specified which
    # runscripts to run. If the file exists load it into the ``test_info`` dictionary,
    # if not, set ``test_info`` as an empty dictionary.
    if os.path.isfile(test_config):
        with open(test_config, "r") as t:
            test_info = yaml.load(t, Loader=yaml.FullLoader)
    else:
        test_info = {}
    logger.debug(test_info)

    # Check if all runscripts need to be tested
    if len(test_info) > 0:
        test_all = False
    else:
        test_all = True

    # Loop through the models in the runscript folder
    for model in os.listdir(runscripts_dir):
        # Check if the users wants to run this model
        if test_all or test_info.get(model, False):
            # Check computer
            model_config = f"{runscripts_dir}/{model}/config.yaml"
            if not os.path.isfile(model_config):
                logger.error(f"'{model_config}' not found!")
            with open(model_config, "r") as c:
                config_test = yaml.load(c, Loader=yaml.FullLoader)
            computers = config_test.get("computers", False)
            if computers:
                # If this runscript is not to be evaluated in this computer, move on
                if info["this_computer"] not in computers:
                    continue
            # Load files to ignore
            ignore_scripts = config_test.get("ignore_scripts", [])
            ignore_scripts.append("config.yaml")

            scripts_info[model] = {}
            # Loop through the testing runscripts of this model
            for script in os.listdir(f"{runscripts_dir}/{model}"):
                # Check if the runscript needs to be tested
                if (
                    test_all
                    or isinstance(test_info.get(model), str)
                    or script in test_info.get(model, [])
                ) and os.path.isfile(f"{runscripts_dir}/{model}/{script}"):
                    # Check that it is actually a runscript
                    if (
                        not any(script in s for s in ignore_scripts)
                        and ".swp" not in script
                    ):
                        # Store information about the runscript
                        scripts_info[model][script.replace(".yaml", "")] = {}
                        scripts_info[model][script.replace(".yaml", "")][
                            "path"
                        ] = f"{runscripts_dir}/{model}/{script}"
                        scripts_info[model][script.replace(".yaml", "")]["state"] = {}
                        ns += 1
            # Delete models that do not contain any script
            if not scripts_info[model]:
                del scripts_info[model]

    scripts_info["general"] = {"num_scripts": ns}

    info["scripts"] = scripts_info

    return info


def read_info_from_rs(info):
    """
    Loads the following information from the runscripts to be tested:
        * ``general.version`` (or ``general.comp_version``)
        * ``general.comp_command`` (if exists)

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info.

    Returns
    -------
    info : dict
        Same as input but but including the new information from the runscripts.
    """
    scripts_info = info["scripts"]
    new_loader = create_env_loader()
    # Loop through the models
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        # Loop through the model's runscripts
        for script, v in scripts.items():
            # Open the script and load variables
            with open(v["path"], "r") as rs:
                runscript = yaml.load(rs, Loader=yaml.SafeLoader)
            v["version"] = runscript["general"].get(
                "comp_version", runscript[model]["version"]
            )
            v["comp_command"] = runscript["general"].get("comp_command", None)
            # Data for iterative coupling
            v["iterative_coupling"] = runscript["general"].get(
                "iterative_coupling", False
            )
            if v["iterative_coupling"]:
                v["nmodels_iterative_coupling"] = 0
                v["iterative_models"] = {}
                for model in runscript:
                    if "model" in model:
                        v["nmodels_iterative_coupling"] += 1
                        # Store the name of the subscripts so that they can be run in
                        # checks
                        v["iterative_models"][model] = {
                            "script": runscript[model]["runscript"]
                        }
            else:
                v["nmodels_iterative_coupling"] = 1

    info["scripts"] = scripts_info

    return info


def del_prev_tests(info):
    """
    Deletes the previews tests that coincide with the ones that are to be run this
    time.

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info.
    """
    scripts_info = info["scripts"]
    user_info = info["user"]
    logger.debug("Deleting previous tests")
    # Loop through the model's runscripts
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        # Loop through the model's runscripts
        for script, v in scripts.items():
            # If the compilation directory for this runscript exists, remove it
            if os.path.isdir(
                f"{user_info['test_dir']}/comp/{model}/{model}-{v['version']}"
            ):
                shutil.rmtree(
                    f"{user_info['test_dir']}/comp/{model}/{model}-{v['version']}"
                )
                if len(os.listdir(f"{user_info['test_dir']}/comp/{model}")) == 0:
                    shutil.rmtree(f"{user_info['test_dir']}/comp/{model}")
            # If the run directory for this runscript exists, remove it
            if os.path.isdir(f"{user_info['test_dir']}/run/{model}/{script}"):
                shutil.rmtree(f"{user_info['test_dir']}/run/{model}/{script}")
                if len(os.listdir(f"{user_info['test_dir']}/run/{model}")) == 0:
                    shutil.rmtree(f"{user_info['test_dir']}/run/{model}")



"""
Backwards compatability for old runscripts
"""
# Python 2 and 3 version agnostic compatiability:
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

# Python Standard Library imports
import logging
import os
import sys
import subprocess

# Third party imports
import six

# Imports from this module
from .yaml_to_dict import *


from esm_rcfile import FUNCTION_PATH


def mini_recursive_run_func(config, func):
    func(config)
    for key, value in six.iteritems(config):
        if isinstance(value, dict):
            if not key == "export_vars":
                mini_recursive_run_func(value, func)


# NOTE: Both of the next to blocks need to be recursive through the whole dictionary...
#
# BLOCK 1: Use mapping table
# Load a mapping table if it exists and swap old and new keys
def remap_old_new_keys(config):
    try:
        mapping_table_old_to_new = "not a thing yet"
        mapping_table = yaml_file_to_dict(mapping_table_old_to_new)
        for script_key, python_key in six.iteritems(mapping_table):
            for config_key, config_value in six.iteritems(config):
                if config_key == script_key:
                    del config[script_key]
                    config[python_key] = config_calue
    except:
        pass


# BLOCK 2: make every key lowercase always
#
# Adapt cases to always be lowercase (I guess this is the new default)
def purify_cases(config):
    import time

    purify_case = True
    if purify_case:
        keys = list(config)
        for key in keys:  # value in six.iteritems(config):
            value = config[key]
            del config[key]
            new_key = key.lower()
            config[new_key] = value
            time.sleep(0.01)


def ShellscriptToUserConfig(runscript_path):
    """
    Generates a User Config from an old Shellscript
    """
    with open(runscript_path) as runscript_file:
        all_lines = runscript_file.readlines()
    hashbang = all_lines[0]
    bad_lines = ("load_all_functions", "general_do_it_all", "#", "set ")
    good_lines = [line.strip() for line in all_lines if not line.startswith(bad_lines)]
    good_lines.insert(0, hashbang)
    good_lines.insert(1, "set -a")
    # Module commands:
    module_commands = [line for line in good_lines if "module" in line]
    # Find index of a command "module purge"
    if "module purge" in module_commands:
        # Anything before module purge is probably irrelevant, so flip the
        # list around first before figuring out which index it is:
        index = module_commands[::-1].index("module purge")
        remaining_module_commands = module_commands[::-1][:index][::-1]
    else:
        remaining_module_commands = module_commands
    module_commands = [l for l in remaining_module_commands if "list" not in l]
    for module_command in module_commands:
        os.system(module_command)
    env_before = dict(os.environ)
    # the next lines remove functions
    all_keys = list(env_before.keys())
    for key in all_keys:
        if "()" in key:
            del env_before[key]
            good_lines.append(
                "unset -f " + key.replace("()", "").replace("BASH_FUNC_", "")
            )

    logging.debug("Got environment from the system %s", env_before)
    with open("cleaned_runscript", "w") as cleaned_runscript:
        for line in good_lines:
            cleaned_runscript.write(line + "\n")
        logging.debug("Finished writing cleaned_runscript")
    command_to_run = "source %s/cleaned_runscript; env" % os.getcwd()
    logging.debug(
        "Using the following command to determine environment in runscript: %s",
        command_to_run,
    )
    pipe1 = subprocess.Popen(command_to_run, stdout=subprocess.PIPE, shell=True)
    output = pipe1.communicate()[0].decode("utf-8")
    logging.debug(output)
    env_after = {}
    for line in output.split("\n"):
        if line and "=" in line:
            key, value = line.split("=", 1)
            if "()" not in key:
                if value:
                    env_after[key] = value
    os.remove("cleaned_runscript")
    diffs = list(set(env_after) - set(env_before))

    known_setups_and_models = os.listdir(FUNCTION_PATH)
    user_config = {}
    logging.debug(diffs)
    solved_diffs = []
    for thisdiff in diffs:
        logging.debug("thisdiff=%s", thisdiff)
        for sim_thing in known_setups_and_models:
            diff_name = thisdiff.replace("_" + sim_thing, "").replace(
                sim_thing + "_", ""
            )
            user_config.setdefault(sim_thing, {})

            if thisdiff.endswith(sim_thing):
                user_config[sim_thing][diff_name] = env_after[thisdiff]
                solved_diffs.append(thisdiff)
                break
            if thisdiff.startswith(sim_thing):
                user_config[sim_thing][diff_name] = env_after[thisdiff]
                solved_diffs.append(thisdiff)
                break
    for k, v in six.iteritems(user_config):
        if v:
            user_config[k] = v
    for solved_diff in solved_diffs:
        diffs.remove(solved_diff)

    solved_diffs = []
    deprecated_diffs = [
        "FUNCTION_PATH",
        "FPATH",
        "machine_name",
        "ESM_USE_C_CALENDAR",
    ]
    user_config["general"] = {}
    for diff in diffs:
        if diff in deprecated_diffs:
            logging.warning(
                "You used a discontinued variable: %s.",
                diff,
            )
        if diff not in deprecated_diffs:
            user_config["general"][diff] = env_after[diff]
            solved_diffs.append(diff)

    for solved_diff in solved_diffs:
        diffs.remove(solved_diff)
    logging.debug("Diffs after removing: %s", diffs)

    # Remove all empty dictionaries:
    for key in list(user_config):
        value = user_config[key]
        if not value:
            del user_config[key]

    # mini_recursive_run_func(user_config, remap_old_new_keys)
    mini_recursive_run_func(user_config, purify_cases)

    return user_config



"""
=====
Usage
=====

This package contains functions to set, get, and use entries stored in the
esmtoolsrc file.

To use ESM RCFile in a project::

    import esm_rcfile

You can set specific values in the ``~/.esmtoolsrc`` with::

    set_rc_entry(key, value)

For example::

    >>> set_rc_entry("SCOPE_CONFIG", "/pf/a/a270077/Code/scope/configs/")

Retriving an entry::

    >>> fpath = get_rc_entry("FUNCTION_PATH")
    >>> print(fpath)
    /pf/a/a270077/Code/esm_tools/esm_tools/configs

With a default value for a non-existing key::

    >>> scope_config = get_rc_entry("SCOPE_CONFIG", "/dev/null")
    >>> print(scope_config)
    /dev/null

Without a default value, you get ``EsmRcfileError``::

    >>> echam_namelist = get_rc_entry("ECHAM_NMLDIR")
    EsmRcFileError: No value for ECHAM_NMLDIR found in esmtoolsrc file!!

This error is also raised if there is no ``~/.esmtoolsrc`` file, and no default
is provided.

You can also get the entire rcfile as a dict::

    >>> rcdict = import_rc_file()


API Documentation
-----------------
"""
import os
import esm_tools

# FIXME(PG): Deprecate this soon. Globals should be in all-caps.
rcfile = RCFILE = os.path.expanduser("~") + "/.esmtoolsrc"


class EsmRcfileError(Exception):
    """Raise this error when there is a problem with ``.esmtoolsrc``."""


class EsmToolsDir(str):
    """
    A string subclass whose instances provide the paths to `esm_tools` folders
    (`configs`, `namelists` and `runscripts`) when evaluated as a string (i.e.
    `str(<your_instance>)` or `<your_instance> + <a_string>`).

    This class is thought as a generalised solution for the problem of removing the
    `FUNCTION_PATH`, `NAMELIST_PATH` and `RUNSCRIPT_PATH` from the `.esmtoolsrc` file,
    and intends to provide those paths correctly, both for virtual environment and open
    runs, right at the time where the variable containing the instance is evaluated
    by the `esm_parser`, as a string.

    This should solve issues with, for example, preprocessing and postprocessing scripts
    called during runtime with a `NONE_YET/<path_to_the_script>`.
    """

    def __init__(self, path_type):
        """
        Initializes the instance. To do that, use `<your_instance_name> =
        EsmToolsDir(<path_type>)` where `path_type` can be::
          - "FUNCTION_PATH"
          - "NAMELIST_PATH"
          - "RUNSCRIPT_PATH"

        Parameters
        ----------
        path_type : str
            ESM-Tools path type to be loaded by the object.
        """
        self.path_type = path_type

    def __str__(self):
        """
        When the instance is evaluated as a string, it returns the correct path to its
        corresponding `esm_tools` folder.

        Returns
        -------
        <esm_tools_folder_type_PATH> : str
            The path to the required folder.
        """
        return self.find_path()

    def __add__(self, add_string):
        """
        When the instance is used together with a sum operation, it returns the correct
        path to its corresponding `esm_tools` folder, plus the operation input.

        Parameters
        ----------
        add_string : str
            String to be added to the path.

        Returns
        -------
        <PATH+add_string> : str
            The path to the required folder plus the string that follows (`add_string`).
        """
        return self.find_path() + add_string

    def find_path(self):
        """
        This method returns the path of the `esm_tools` folder required.

        Returns
        -------
        <esm_tools_folder_type_PATH> : str
            The path to the required folder.
        """
        if self.path_type == "FUNCTION_PATH":
            cpath = esm_tools.get_config_filepath(".")
            if cpath:
                return f"{cpath}/"
            else:
                return "/dev/null"
        if self.path_type == "NAMELIST_PATH":
            npath = esm_tools.get_namelist_filepath(".")
            if npath:
                return f"{npath}/"
            else:
                return "/dev/null"
        if self.path_type == "RUNSCRIPT_PATH":
            rpath = esm_tools.get_runscript_filepath(".")
            if rpath:
                return f"{rpath}/"
            else:
                return "/dev/null"
        raise Exception("Incorrect path type!")


def set_rc_entry(key, value):
    """
    Sets values in ``esmtoolsrc``

    Parameters
    ----------
    key : str
    value : str

    Note
    ----
    Using this functions modifies the ``rcfile``; which is stored in the
    current user's home directory.
    """
    all_lines = [key + "=" + value]

    if os.path.isfile(RCFILE):
        with open(RCFILE) as rc:
            for line in rc.readlines():
                line = line.strip()
                if not key == line.split("=", 1)[0]:
                    all_lines.append(line)
        os.remove(RCFILE)

    with open(RCFILE, "w") as rc:
        for line in all_lines:
            rc.write(line + "\n")


def get_rc_entry(key, default=None):
    """
    Gets a specific entry

    Parameters
    ----------
    key : str
    default : str

    Returns
    -------
    str
        Value for key, or default if provided

    Raises
    ------
    EsmRcfileError
        * Raised if key cannot be found in the rcfile and no default is
          provided
        * Raised if the esmtoolsrc file cannot be found and no default is
          provided.
    """
    if os.path.isfile(RCFILE):
        with open(RCFILE) as rc:
            for line in rc.readlines():
                line = line.strip()
                if line.split("=", 1)[0] == key.upper():
                    return line.split("=", 1)[1]
            if default:
                return default
            else:
                raise EsmRcfileError("No value for %s found in esmtoolsrc file!" % key)
    if default:
        return default
    else:
        raise EsmRcfileError("The file esmtoolsrc file was not found!")


def import_rc_file():
    """
    Gets current values of the esmtoolsrc file

    Returns
    -------
    dict
        A dictionary representation of the rcfile
    """
    if os.path.isfile(RCFILE):
        rcdict = {}
        with open(RCFILE) as rc:
            for line in rc.readlines():
                line = line.strip()
                rcdict[line.split("=", 1)[0]] = line.split("=", 1)[1]
        return rcdict
    raise EsmRcfileError("The file esmtoolsrc file was not found!")


# PG: Should this be in a if __name__ == "__main__" ?
# MA: The following lines are most likely never reached since the work
# from PG in implementing the venv. It is probably wise to remove them
# in the next cleanup.
if os.path.isfile(RCFILE):
    FUNCTION_PATH = get_rc_entry("FUNCTION_PATH", "NONE_YET")
else:
    FUNCTION_PATH = "NONE_YET"



#!/usr/bin/env python
"""
A small wrapper that combines the shell interface and the Python interface
"""
from . import event_handlers

event_handlers.signal_listener()

# Import from Python Standard Library
import argparse
import logging
import os
import sys

from esm_motd import check_all_esm_packages
from loguru import logger

from .helpers import SmartSink
from .sim_objects import *


def parse_shargs():
    """The arg parser for interactive use"""
    parser = argparse.ArgumentParser()
    parser.add_argument("runscript", default=None)

    parser.add_argument(
        "-d",
        "--debug",
        help="Print lots of debugging statements",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.ERROR,
    )

    parser.add_argument(
        "-v",
        "--verbose",
        help="Be verbose",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--contained-run",
        help="Run in a virtual environment",
        action="store_true",
        default=None,
    )

    parser.add_argument(
        "--open-run",
        help="Run in default install (not in virtual environment)",
        action="store_true",
        default=None,
    )

    parser.add_argument(
        "-e", "--expid", help="The experiment ID to use", default="test"
    )

    parser.add_argument(
        "-c",
        "--check",
        help="Run in check mode (don't submit job to supercomputer)",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "-P",
        "--profile",
        help="Write profiling information (esm-tools)",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--modify-config",
        "-m",
        dest="modify",
        help="[m]odify configuration",
        default="",  # kh 15.07.20 "usermods.yaml"
    )

    # parser.add_argument(
    #    "-j",
    #    "--last_jobtype",
    #    help="Write the jobtype this run was called from (esm-tools internal)",
    #    default="command_line",
    # )
    #
    parser.add_argument(
        "-t",
        "--task",
        help="The task to run. Choose from: prepcompute, post, couple, tidy",
        default="unknown",
    )

    parser.add_argument(
        "-i",
        "--inspect",
        help="Show some information, choose a keyword from 'overview', 'namelists'",
        default=None,
    )

    parser.add_argument(
        "-p",
        "--pid",
        help="The PID of the task to observe.",
        default=-666,
    )

    parser.add_argument(
        "-s",
        "--start_date",
        help="The start_date of the run, overwriting settings in the date file.",
        default=None,
    )

    parser.add_argument("-x", "--exclude", help="e[x]clude this step", default=None)
    parser.add_argument("-o", "--only", help="[o]nly do this step", default=None)
    parser.add_argument(
        "-r",
        "--run_number",
        help="run_number for this run, overwriting settings in date file",
        default=None,
    )

    # PG: Might not work anymore:
    parser.add_argument(
        "-U",
        "--update",
        help="[U]date the tools from the current version and the runscript",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--update-filetypes",
        help="Updates the requested files from external sources in a currently ongoing "
        "simulation. We strongly advise against using this option unless you "
        "really know what you are doing.",
        nargs="+",
        default=[],
    )

    parser.add_argument(
        "--no-motd",
        help="supress the printing of MOTD",
        default=False,
        action="store_true",
    )

    return parser.parse_args()


def main():

    ARGS = parse_shargs()

    check = False
    profile = False
    update = False
    expid = "test"
    pid = -666
    start_date = None
    run_number = None
    jobtype = "unknown"
    verbose = False
    inspect = None
    use_venv = None
    modify_config_file = None
    no_motd = False

    parsed_args = vars(ARGS)

    original_command = " ".join(sys.argv[1:])

    if "check" in parsed_args:
        check = parsed_args["check"]
    if "profile" in parsed_args:
        profile = parsed_args["profile"]
    if "pid" in parsed_args:
        pid = parsed_args["pid"]
    if "start_date" in parsed_args:
        start_date = parsed_args["start_date"]
    if "run_number" in parsed_args:
        run_number = parsed_args["run_number"]
    if "update" in parsed_args:
        update = parsed_args["update"]
    if "update_filetypes" in parsed_args:
        update_filetypes = parsed_args["update_filetypes"]
    if "expid" in parsed_args:
        expid = parsed_args["expid"]
    if "task" in parsed_args:
        jobtype = parsed_args["task"]
    if "verbose" in parsed_args:
        verbose = parsed_args["verbose"]
    if "inspect" in parsed_args:
        inspect = parsed_args["inspect"]
    if parsed_args["contained_run"] and parsed_args["open_run"]:
        print("You have set both --contained-run and --open-run, this makes no sense.")
        print(parsed_args)
        sys.exit(1)
    if parsed_args["contained_run"] is not None:
        use_venv = parsed_args["contained_run"]
    if parsed_args["open_run"] is not None:
        use_venv = not parsed_args["open_run"]
    if "modify" in parsed_args:
        modify_config_file = parsed_args["modify"]
    if "no_motd" in parsed_args:
        no_motd = parsed_args["no_motd"]

    command_line_config = {}
    command_line_config["check"] = check
    command_line_config["profile"] = profile
    command_line_config["update"] = update
    command_line_config["update_filetypes"] = update_filetypes
    command_line_config["expid"] = expid
    command_line_config["launcher_pid"] = pid
    command_line_config["current_date"] = start_date
    command_line_config["run_number"] = run_number
    command_line_config["jobtype"] = jobtype
    # command_line_config["last_jobtype"] = ARGS.last_jobtype
    command_line_config["verbose"] = verbose
    command_line_config["inspect"] = inspect
    command_line_config["use_venv"] = use_venv
    command_line_config["no_motd"] = no_motd
    if modify_config_file:
        command_line_config["modify_config_file"] = modify_config_file

    # runscript_from_cmdline = filter(lambda x: x.endswith(".yaml"), sys.argv)
    # runscript_from_cmdline = list(runscript_from_cmdline)[0]
    # runscript_full_path = os.path.realpath(runscript_from_cmdline)
    runscript_full_path = os.path.realpath(ARGS.runscript)
    runscript_dir, runscript = os.path.split(runscript_full_path)
    runscript_dir += "/"

    # this might contain the relative path but it will be taken care of later
    command_line_config["original_command"] = original_command.strip()
    command_line_config["started_from"] = runscript_dir

    # only the yaml file, without the path
    command_line_config["scriptname"] = runscript
    # full path including the yaml file: runscript_dir + runscript
    command_line_config["runscript_abspath"] = runscript_full_path

    # Define a sink object to store the logs. Path of the logs can be later specified
    # by using <sink_obj>.def_path(<path>)
    trace_sink = SmartSink()
    logger.trace_sink = trace_sink

    logger.remove()
    logger.add(trace_sink.sink, level="TRACE")

    logger.add(sys.stdout, level="INFO", format="{message}")
    if verbose:
        logger.debug(f"Started from: {command_line_config['started_from']}")
        logger.debug(f"starting (jobtype): {jobtype}")
        logger.debug(command_line_config)

    Setup = SimulationSetup(command_line_config)
    # if not Setup.config['general']['submitted']:
    if not Setup.config["general"]["submitted"] and not no_motd:
        check_all_esm_packages()
    Setup()



import datetime
import os
import site

import questionary

import pathlib
import subprocess
import sys
import venv

esm_tools_modules = [
    "esm_calendar",
    "esm_database",
    "esm_environment",
    "esm_master",
    "esm_parser",
    "esm_rcfile",
    "esm_runscripts",
    "esm_tools",
    "esm_plugin_manager",
    "esm_version_checker",
]


def get_base_prefix_compat():
    """Get base/real prefix, or sys.prefix if there is none."""
    return (
        getattr(sys, "base_prefix", None)
        or getattr(sys, "real_prefix", None)
        or sys.prefix
    )


def in_virtualenv():
    return get_base_prefix_compat() != sys.prefix


class _EnvBuilder(venv.EnvBuilder):
    def __init__(self, *args, **kwargs):
        self.context = None
        super().__init__(*args, **kwargs)

    def post_setup(self, context):
        self.context = context


def _venv_create(venv_path):
    venv_builder = _EnvBuilder(with_pip=True)
    venv_builder.create(venv_path)
    return venv_builder.context


def _run_python_in_venv(venv_context, command):
    command = [venv_context.env_exe] + command
    return subprocess.check_call(command)


def _run_bin_in_venv(venv_context, command):
    command[0] = str(pathlib.Path(venv_context.bin_path).joinpath(command[0]))
    return subprocess.check_call(command)


def _source_and_run_bin_in_venv(venv_context, command, shell):
    source_command = " ".join(
        ["source", venv_context.bin_path + "/activate", "&&", " "]
    )
    command = source_command + command
    return subprocess.check_call(command, shell=shell)


def _install_tools(venv_context, config):
    """
    Installs the ESM-Tools packages for a virtual environment, taking into account
    the user's specifications for editable packages and desired branches.

    To control which packages are installed in editable mode the user can add the
    following to their runscript:

    .. code-block:: yaml

       general:
           install_<esm_package>_editable: True/False

    To control which package branch is installed (compatible with editable mode and
    non-editable mode) the user can add the following to their runscript:

    .. code-block:: yaml

       general:
           install_<esm_package>_branch: <branch_name>

    Parameters
    ----------
    venv_context : type
        Some description
    config : dict
        Configuration dictionary for this run
    """

    # First installation of all packages, with the desired mode (editable/non-editable),
    # and branches, together with all their dependencies
    _install_tools_general(venv_context, config)
    # Some packages, such as `esm_tools` have other packages as dependencies (i.e.
    # `esm_parser`). In the previous step, if a package is installed for which a
    # dependency is editable and/or branched, this dependency gets back to non-editable
    # realese-branch. The following line resinstalls all the editable/branched packages
    # again, this time without dependencies.
    _install_tools_general(venv_context, config, deps=False)


def _install_tools_general(venv_context, config, deps=True):
    """
    Actual installer of ESM-Tools packages for virtual environments. Used by
    `_install_tools` method to correctly install packages with the user's requested
    options for each package (editable/non-editable and branch). See `_install_tools`
    documentation for more information.

    Parameters
    ----------
    venv_context : type
        Some description
    config : dict
        Configuration dictionary for this run
    deps : bool
        Boolean indicating whether dependencies should be installed or not
    """
    # Setup the --no-deps flag if necessary
    if not deps:
        no_deps_flag = ["--no-deps"]
    else:
        no_deps_flag = []
    # Loop through the esm_tools packages to be installed
    for tool in esm_tools_modules:
        # Module info (url, editable install, branch...)
        url = f"https://github.com/esm-tools/{tool}"
        user_wants_editable = config["general"].get(f"install_{tool}_editable", False)
        user_wants_branch = config["general"].get(f"install_{tool}_branch")
        # If the package is editable install it in <EXP_PATH>/src/esm-tools/
        if user_wants_editable:
            # Make sure the directory exists:
            src_dir = pathlib.Path(
                config["general"]["experiment_dir"] + f"/src/esm-tools/{tool}"
            )
            if not src_dir.exists():
                src_dir.mkdir(parents=True, exist_ok=True)
            # Select branch if necessary
            if user_wants_branch:
                branch_command = f" -b {user_wants_branch} "
            else:
                branch_command = ""
            # Clone from git
            if deps:
                subprocess.check_call(
                    f"git clone --quiet {branch_command} {url} {src_dir}", shell=True
                )
            # Carry out the editable installation (with or without dependencies)
            _run_bin_in_venv(
                venv_context,
                [
                    "pip",
                    "install",
                    "-q",
                    f"--find-links={os.environ.get('HOME')}/.cache/pip/wheels",
                    "-e",
                    src_dir,
                ]
                + no_deps_flag,
            )
            _run_bin_in_venv(
                venv_context,
                [
                    "pip",
                    "wheel",
                    "-q",
                    f"--wheel-dir={os.environ.get('HOME')}/.cache/pip/wheels",
                    "-e",
                    src_dir,
                ]
                + no_deps_flag,
            )
        # If the package is not editable then do a standard installation.
        # Note: this step only runs with the `--no-deps` flag if the user has specified
        # a branch, as this flags means also that is the second time passing through
        # here, and we don't want to waste time on installing everything a second time
        # if not necessary.
        elif deps or (not deps and user_wants_branch):
            url = f"git+{url}"
            if user_wants_branch:
                url += f"@{user_wants_branch}"
            # NOTE(PG): We need the -U flag to ensure the branch is actually installed.
            _run_bin_in_venv(
                venv_context,
                [
                    "pip",
                    "install",
                    "-q",
                    f"--find-links={os.environ.get('HOME')}/.cache/pip/wheels",
                    "-U",
                    url,
                ]
                + no_deps_flag,
            )
            _run_bin_in_venv(
                venv_context,
                [
                    "pip",
                    "wheel",
                    "-q",
                    f"--wheel-dir={os.environ.get('HOME')}/.cache/pip/wheels",
                    url,
                ]
                + no_deps_flag,
            )


def _install_required_plugins(venv_context, config):
    required_plugins = []
    for sub_cfg_key, sub_cfg in config.items():
        if isinstance(sub_cfg, dict):
            if "required_plugins" in sub_cfg:
                try:
                    assert isinstance(sub_cfg["required_plugins"], list)
                except AssertionError:
                    print(f"ERROR -- required plugins in {sub_cfg_key} must be a list!")
                    sys.exit(1)
                required_plugins += sub_cfg["required_plugins"]
    for required_plugin in required_plugins:
        _run_bin_in_venv(venv_context, ["pip", "install", "-q", required_plugin])


def venv_bootstrap(config):
    """Bootstraps your run into a virtual environment"""
    if not config["general"].get("use_venv"):
        if (
            config["general"].get("use_venv") is None
            and config["general"]["command_line_config"]["use_venv"] is None
        ):
            config = _integorate_user_venv(config)
            config["general"]["command_line_config"]["use_venv"] = config["general"][
                "use_venv"
            ]
    if config["general"].get("use_venv", False):
        if not in_virtualenv():
            venv_path = pathlib.Path(config["general"]["experiment_dir"]).joinpath(
                ".venv_esmtools"
            )
            if venv_path.exists():
                print(f"{venv_path} already exists, reusing...")
                venv_context = _EnvBuilder(with_pip=True).ensure_directories(venv_path)
            else:
                print(
                    f"Building virtual env, please be patient (this takes about 3 minutes)..."
                )
                start_time = datetime.datetime.now()
                venv_path = pathlib.Path(config["general"]["experiment_dir"]).joinpath(
                    ".venv_esmtools"
                )
                venv_context = _venv_create(venv_path)
                _run_python_in_venv(
                    venv_context, ["-m", "pip", "-q", "install", "-U", "pip"]
                )
                _run_python_in_venv(
                    venv_context, ["-m", "pip", "-q", "install", "-U", "wheel"]
                )
                _install_tools(venv_context, config)
                _install_required_plugins(venv_context, config)
                print(
                    f"...finished {datetime.datetime.now() - start_time}, restarting your job in the virtual env"
                )
            sys.argv[0] = pathlib.Path(sys.argv[0]).name
            # NOTE(PG): This next line allows the job to restart itself in the
            # virtual environment.
            _source_and_run_bin_in_venv(
                venv_context, " ".join(sys.argv) + " --contained-run", shell=True
            )
            sys.exit(0)
    if not config["general"].get("use_venv"):
        if "--open-run" not in config["general"]["original_command"]:
            config["general"]["original_command"] += " --open-run"
        if (
            "--open-run"
            not in config["general"]["command_line_config"]["original_command"]
        ):
            config["general"]["command_line_config"][
                "original_command"
            ] += " --open_run"
    return config


def find_package(pkg):
    pool_root = os.path.abspath(site.getsitepackages()[0] + f"/{pkg}")
    if not os.path.exists(pool_root):
        # Check if an egg link exists:
        if os.path.isfile(site.getusersitepackages() + f"/{pkg}.egg-link"):
            with open(site.getusersitepackages() + f"/{pkg}.egg-link") as egg_file:
                egg_contents = egg_file.readlines()[0].strip()
            pool_root = egg_contents + f"/{pkg}"
        else:
            pool_root = os.path.abspath(site.getusersitepackages() + f"{pkg}")
    return pool_root


def _integorate_user_venv(config):
    questionary.print("\t" + 100 * "=")
    questionary.print(
        "\t\tRunning jobs can optionally be encapsulated into a virtual environment\n"
    )
    questionary.print(
        "\t\tThis shields the run from changes made to the remainder of the ESM-Tool installation\n"
    )
    questionary.print(
        "\t\tBefore the first run, a local copy will be installed in the experiment tree,"
    )
    questionary.print("\t\tand **this** installation will be used:")
    questionary.print(
        f"\t\t\t {config['general']['experiment_dir']}/.venv_esmtools/lib/python-{sys.version_info.major}.{sys.version_info.minor}/site-packages/esm-tools"
    )
    questionary.print("\t\tinstead of:")
    questionary.print(f"\t\t\t{find_package('esm-tools')}\n")
    questionary.print(
        "\n\t\tIf you choose to use a virtual environment, a folder named `.venv_esmtools`"
    )
    questionary.print(
        "\t\twill be created at the root of your experiment. This contains all the Python"
    )
    questionary.print(
        "\t\tlibraries used by the esm-tools. The first installation induces some overhead (~2-3 minutes)."
    )
    questionary.print(
        "\n\t\tYou may also select to install some packages in 'editable mode', in which case"
    )
    questionary.print(
        "\t\tthey will be installed in a folder `src/esm_tools/<package_name>` in the root of"
    )
    questionary.print(
        "\t\tyour experiment. Any changes made to code in that folder **will** influence how the"
    )
    questionary.print("\t\tesm-tools behave.")
    questionary.print("\n\t\tNote regarding config yamls and namelists")
    questionary.print("\t\t-----------------------------------------")
    questionary.print(
        "\t\tWhen using a virtual environment, config files and namelists will come of the"
    )
    questionary.print(
        "\t\tfolder .venv_esmtools listed above and **not** from your user install directory."
    )
    questionary.print(
        "\t\tYou should make **all** changes to the namelists and config files via your user"
    )
    questionary.print("\t\trunscript. This is recommended in all cases!!!")
    questionary.print(
        "\t\t --> Changing any settings in your user install ($HOME or similar) will therefore not change your run!! <--",
        style="fg:red",
    )
    questionary.print("\n\t\tNotes from the development team:")
    questionary.print("\t\t--------------------------------")
    questionary.print(
        "\n\t\t --> Using an encapsulated environment is recommended for production runs <--\n",
        style="fg:green",
    )
    questionary.print("\t Happy simulating!")
    questionary.print(
        "\t Dirk Barbi, Paul Gierz, Nadine Wieters, Miguel Andrés-Martínez, Deniz Ural, and the rest of the development team."
    )
    questionary.print("\t" + 100 * "=")

    user_confirmed = False
    while not user_confirmed:
        response = questionary.select(
            "What do you want to do?",
            choices=[
                "Run in virtualenv (You may set the flag `--contained-run` during your run call or set `general.use_venv: True`)",
                "Run using default installation (You may set the flag `--open-run` during your run call or set `general.use_venv: False`)",
                "Quit right now to adapt your runscript",
            ],
        ).ask()  # returns value of selection
        if "Quit" in response:
            sys.exit(0)
        config["general"]["use_venv"] = "Run in virtualenv" in response
        user_confirmed = questionary.confirm("Are you sure?").ask()
    if "Run in virtualenv" in response:
        editable_mode_choices = questionary.checkbox(
            "Which of the following packages do you want to install in 'editable mode'? End with <Enter>",
            choices=esm_tools_modules,
        ).ask()
        for choice in editable_mode_choices:
            config["general"][f"install_{choice}_editable"] = True
    return config


if __name__ == "__main__":
    dummy_config = {"general": {"experiment_dir": "/foo/bar"}}
    config = _integorate_user_venv(dummy_config)
    print(config)



"""Console script for esm_master."""
import argparse
import sys

# import logging
# logging.basicConfig(level=logging.DEBUG)
from . import __version__
from esm_motd import check_all_esm_packages
from .esm_master import main_flow


def main():

    parser = argparse.ArgumentParser(
        prog="esm_master",
        description="tool for downloading, configuring and compiling.",
    )
    parser.add_argument(
        "target",
        metavar="target",
        nargs="?",
        type=str,
        help="name of the target (leave empty for full list of targets)",
    )
    parser.add_argument(
        "--check",
        "-c",
        action="store_true",
        default=False,
        help="show what would be done, not doing anything",
    )
    parser.add_argument(
        "--verbose", "-v", action="count", default=False, help="toggle verbose mode"
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )

    parser.add_argument(
        "--modify-config",
        "-m",
        dest="modify",
        help="[m]odify configuration",
        default="",  # kh 15.07.20 "usermods.yaml"
    )

    # kh 21.07.20
    parser.add_argument(
        "--ignore-errors",
        "-i",
        dest="ignore",
        help="Ignore errors",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--keep-task-script",
        "-k",
        dest="keep",
        action="store_true",
        default=False,
        help="Keep shell script generated to perform compilation/configuration jobs",
    )
    parser.add_argument("--generate_tab_complete", action="store_true")
    parser.add_argument("--list_all_targets", action="store_true")

    parser.add_argument(
        "--no-motd",
        help="supress the printing of MOTD",
        default=False,
        action="store_true",
    )

    parsed_args = vars(parser.parse_args())

    target = ""
    check = False
    verbose = False
    modify_config_file = False
    no_motd = False

    if parsed_args:
        if "target" in parsed_args:
            target = parsed_args["target"]
        if "check" in parsed_args:
            check = parsed_args["check"]
        if "verbose" in parsed_args:
            verbose = parsed_args["verbose"]
        if "keep" in parsed_args:
            keep = parsed_args["keep"]
        if "modify" in parsed_args:
            modify_config_file = parsed_args["modify"]
        if "no_motd" in parsed_args:
            no_motd = parsed_args["no_motd"]

    if not target:
        target = ""

    if not no_motd:
        check_all_esm_packages()

    main_flow(parsed_args, target)


if __name__ == "__main__":
    main(sys.argv[1:])



import copy
import logging
import os
import sys

import esm_parser
from esm_calendar import Calendar, Date

from . import batch_system, helpers


def run_job(config):
    helpers.evaluate(config, "prepare", "prepare_recipe")
    return config


def mini_resolve_variable_date_file(date_file, config):
    while "${" in date_file:
        pre, post = date_file.split("${", 1)
        variable, post = post.split("}", 1)
        if "." in variable:
            variable_section, variable = variable.split(".")
            answer = config[variable_section].get(variable)
        else:
            answer = config["general"].get(variable)
            if not answer:
                answer = config.get("env", {}).get(variable)
                if not answer:
                    try:
                        assert variable.startswith("env.") or variable.startswith(
                            "general."
                        )
                    except AssertionError:
                        print(
                            "The date file contains a variable which is not in the >>env<< or >>general<< section. This is not allowed!"
                        )
                        print(f"date_file = {date_file}")
                        sys.exit(1)
        date_file = pre + answer + post
    return date_file


def _read_date_file(config):
    import logging
    import os

    date_file = (
        f"{config['general']['experiment_dir']}/scripts"
        f"/{config['general']['expid']}_{config['general']['setup_name']}.date"
    )

    date_file = mini_resolve_variable_date_file(date_file, config)

    if os.path.isfile(date_file):
        logging.info("Date file read from %s", date_file)
        with open(date_file) as date_file:
            date, run_number = date_file.readline().strip().split()
            run_number = int(run_number)
        write_file = False
    else:
        logging.info("No date file found %s", date_file)
        logging.info("Initializing run_number=1 and date=18500101")
        date = config["general"].get("initial_date", "18500101")
        run_number = 1
        write_file = True
    config["general"]["run_number"] = run_number
    config["general"]["current_date"] = date
    logging.info("current_date = %s", date)
    logging.info("run_number = %s", run_number)
    return config


def check_model_lresume(config):
    if config["general"]["run_number"] != 1:
        for model in config["general"]["valid_model_names"]:
            config[model]["lresume"] = True
    else:
        # Did the user give a value? If yes, keep it, if not, first run:
        for model in config["general"]["valid_model_names"]:
            if "lresume" in config[model]:
                user_lresume = config[model]["lresume"]
            else:
                user_lresume = False

            if isinstance(user_lresume, str) and "${" in user_lresume:
                user_lresume = esm_parser.find_variable(
                    model, user_lresume, config, [], []
                )
            if type(user_lresume) == str:

                if user_lresume == "0" or user_lresume.upper() == "FALSE":
                    user_lresume = False
                elif user_lresume == "1" or user_lresume.upper() == "TRUE":
                    user_lresume = True
            elif isinstance(user_lresume, int):
                if user_lresume == 0:
                    user_lresume = False
                elif user_lresume == 1:
                    user_lresume = True
            config[model]["lresume"] = user_lresume
    for model in config["general"]["valid_model_names"]:
        # Check if lresume contains a variable which might be set in a different model, and resolve this case
        if (
            "lresume" in config[model]
            and isinstance(config[model]["lresume"], str)
            and "${" in config[model]["lresume"]
        ):
            lr = esm_parser.find_variable(
                model, config[model]["lresume"], config, [], []
            )
            config[model]["lresume"] = eval(lr)
    return config


def resolve_some_choose_blocks(config):
    # from esm_parser import choose_blocks

    # Component-specific environment variables into ``computer``
    # before ``computer`` ``choose_`` blocks are resolved
    model_env_into_computer(config)

    esm_parser.choose_blocks(config, blackdict=config._blackdict)
    return config


def model_env_into_computer(config):
    """
    This function allows to store in the ``computer`` dictionary, variables that were
    defined inside ``environment_changes`` or ``compile/runtime_environment_changes``
    in the components.

    It excludes ``module_actions`` and ``export_vars`` dictionaries as those are
    resolved later.

    This function is necessary for controlling ``choose_`` blocks in the computer file
    from the component configuration file (i.e. add ``useMPI`` case to the component to
    control which ``useMPI`` case is selected in the computer file).

    This function works both for compilation time and run time, and the result is that
    all components work under the same environment. The only exception is for
    the compilation of components, where ``add_export_vars`` and ``add_module_actions``
    are excluded from the merging into ``computer``, and are included individually in
    respective compilation scripts.

    Later on, it might be desirable to always split the environments both for compiling
    (done by Paul Gierz, but this function would need to be adapted) and running (not
    done yet).

    If this script gives you problems contact Miguel Andres-Martinez
    (miguel.andres-martinez@awi.de).

    Parameters
    ----------
    config : dict
        Dictionary containing the simulation/compilation information

    Raises
    ------
    User Note/Error
        If the same variable is found in two or more different component environments.
        Asks the user how to proceed.
    """

    # import logging
    # from esm_parser import basic_choose_blocks, dict_merge, user_note, user_error, pprint_config

    # Get which type of changes are to be applied to the environment
    run_or_compile = config.get("general", {}).get("run_or_compile", "runtime")
    thesechanges = run_or_compile + "_environment_changes"
    # List the component names
    models = config.get("general", {}).get("models", [])
    # ``env_vars`` stores information about the environment changes to inform
    # about conflicts
    env_vars = {}
    # Loop through the models
    for model in models:
        # Update the models ``environment_changes`` with the ``compiletime/runtime_
        # environment_changes
        modelconfig = copy.deepcopy(config[model])
        modelconfig["environment_changes"] = modelconfig.get("environment_changes", {})
        if thesechanges in modelconfig:
            if "environment_changes" in modelconfig:
                modelconfig["environment_changes"].update(modelconfig[thesechanges])
            else:
                modelconfig["environment_changes"] = modelconfig[thesechanges]
        # Resolve ``choose_`` blocks, ``add_`` and ``remove_`` inside ``environment_
        # changes``
        esm_parser.basic_choose_blocks(modelconfig["environment_changes"], config)
        # Set to true when specified by the user in ``env_overwrite`` or when this
        # method has been already called once in this run
        overwrite = config[model].get("env_overwrite", False)
        # Loop through every variable in ``environment_changes`` except ``export_vars``
        # and ``module_actions`` as those are solved by ``esm_environment`` later and
        # need of the solving of later ``choose_`` blocks.
        for key, value in modelconfig["environment_changes"].items():
            if (
                key
                not in [
                    "export_vars",
                    "module_actions",
                    "add_export_vars",
                    "add_module_actions",
                    "unset_vars",
                    "add_unset_vars",
                ]
                and "computer" in config
                and not overwrite
                # and run_or_compile=="runtime"
            ):
                # If the key is already included in ``env_vars``, the key variable has
                # been already modified by a previous model and a warning needs to be
                # raised. Do this check only on the first step (user interaction step).
                if key in env_vars and config["general"]["run_number"] == 1:
                    # Previous model with the same key
                    model0 = env_vars[key][1]
                    while True:
                        # Warn the user about the overwriting of the variable
                        user_note("Environment conflict", f"In '{model0}':")
                        pprint_config({key: env_vars[key][0]})
                        logging.info("\nIn '" + model + "':")
                        pprint_config({key: value})
                        # Ask the user how to proceed if it is not a `tidy job
                        if not config["general"]["jobtype"] == "tidy":
                            user_answer = input(
                                f"Environment variable '{key}' defined in '{model0}' is "
                                + "going to be overwritten by the one defined in "
                                + f"'{model}'. Are you okay with that? (Y/n): "
                            )
                        # If it is a ``tidy`` job, the user has already
                        # interacted and accepted the overwriting.
                        else:
                            user_answer = "Y"
                            logging.info(
                                f"Environment variable '{key}' defined in '{model0}' is "
                                + f"overwritten by the one defined in '{model}'."
                            )
                        # If the user selects ``Y``, overwrite the environment variable
                        if user_answer == "Y":
                            config[model]["env_overwrite"] = True
                            break
                        # If the user selects ``n`` raise a user error with recommendations
                        elif user_answer == "n":
                            config[model]["env_overwrite"] = False
                            user_error(
                                "Environment conflict",
                                "You were not happy with the environment variable "
                                + f"'{key}' in '{model0}' being overwritten by the same "
                                + f"variable in '{model}'. If you are running a "
                                + "coupled setup, we recommend that you resolve this "
                                + "conflict inside the coupled setup file, by "
                                + "specifying unconflicting environments for each "
                                + "model, or contact the setup developers.",
                            )
                        else:
                            logging.info("Wrong answer, please choose Y/n.")
                # Merge variable into the ``computer`` dictionary so that it becomes
                # part of the general environment.
                esm_parser.dict_merge(config["computer"], {key: value})
                # Add the variable to ``env_vars`` so it can be checked for conflicts
                # with other models.
                env_vars[key] = [value, model]


def _initialize_calendar(config):
    config = set_restart_chunk(config)
    config = set_leapyear(config)
    config = set_overall_calendar(config)
    if config["general"]["reset_calendar_to_last"]:
        config = find_last_prepared_run(config)
    config = set_most_dates(config)
    if not "iterative_coupling" in config["general"]:
        config["general"]["chunk_number"] = 1

        if config["general"]["run_number"] == 1:
            config["general"]["first_run_in_chunk"] = True
        else:
            config["general"]["first_run_in_chunk"] = False
        if config["general"]["next_date"] >= config["general"]["final_date"]:
            config["general"]["last_run_in_chunk"] = True
        else:
            config["general"]["last_run_in_chunk"] = False
    return config


def set_restart_chunk(config):
    nyear, nmonth, nday, nhour, nminute, nsecond = 0, 0, 0, 0, 0, 0
    nyear = int(config["general"].get("nyear", nyear))
    if not nyear:
        nmonth = int(config["general"].get("nmonth", nmonth))
    if not nyear and not nmonth:
        nday = int(config["general"].get("nday", nday))
    if not nyear and not nmonth and not nday:
        nhour = int(config["general"].get("nhour", nhour))
    if not nyear and not nmonth and not nday and not nhour:
        nminute = int(config["general"].get("nminute", nminute))
    if not nyear and not nmonth and not nday and not nhour and not nminute:
        nsecond = int(config["general"].get("nsecond", nsecond))
    if (
        not nyear
        and not nmonth
        and not nday
        and not nhour
        and not nminute
        and not nsecond
    ):
        nyear = 1
    config["general"]["nyear"] = nyear
    config["general"]["nmonth"] = nmonth
    config["general"]["nday"] = nday
    config["general"]["nhour"] = nhour
    config["general"]["nminute"] = nminute
    config["general"]["nsecond"] = nsecond
    return config


def set_leapyear(config):
    # make sure all models agree on leapyear
    if "leapyear" in config["general"]:
        for model in config["general"]["valid_model_names"]:
            config[model]["leapyear"] = config["general"]["leapyear"]
    else:
        for model in config["general"]["valid_model_names"]:
            if "leapyear" in config[model]:
                for other_model in config["general"]["valid_model_names"]:
                    if "leapyear" in config[other_model]:
                        if (
                            not config[other_model]["leapyear"]
                            == config[model]["leapyear"]
                        ):
                            print(
                                "Models "
                                + model
                                + " and "
                                + other_model
                                + " do not agree on leapyear. Stopping."
                            )
                            sys.exit(43)
                    else:
                        config[other_model]["leapyear"] = config[model]["leapyear"]
                config["general"]["leapyear"] = config[model]["leapyear"]
                break

    if not "leapyear" in config["general"]:
        for model in config["general"]["valid_model_names"]:
            config[model]["leapyear"] = True
        config["general"]["leapyear"] = True
    return config


def set_overall_calendar(config):

    # set the overall calendar
    if config["general"]["leapyear"]:
        config["general"]["calendar"] = Calendar(1)
    else:
        config["general"]["calendar"] = Calendar(0)
    return config


def find_last_prepared_run(config):

    calendar = config["general"]["calendar"]
    current_date = Date(config["general"]["current_date"], calendar)
    initial_date = Date(config["general"]["initial_date"], calendar)
    delta_date = (
        config["general"]["nyear"],
        config["general"]["nmonth"],
        config["general"]["nday"],
        config["general"]["nhour"],
        config["general"]["nminute"],
        config["general"]["nsecond"],
    )

    while True:
        if current_date < initial_date:
            break

        next_date = current_date.add(delta_date)
        end_date = next_date - (0, 0, 1, 0, 0, 0)

        datestamp = (
            current_date.format(form=9, givenph=False, givenpm=False, givenps=False)
            + "-"
            + end_date.format(form=9, givenph=False, givenpm=False, givenps=False)
        )

        # Solve base_dir with variables
        base_dir = esm_parser.find_variable(
            ["general", "base_dir"], config["general"]["base_dir"], config, [], True
        )

        if os.path.isdir(
            base_dir + "/" + config["general"]["expid"] + "/run_" + datestamp
        ):
            config["general"]["current_date"] = current_date
            return config

        current_date = current_date - delta_date

    print("ERROR: Could not find a prepared run.")
    sys.exit(42)


def set_most_dates(config):

    calendar = config["general"]["calendar"]
    if isinstance(config["general"]["current_date"], Date):
        current_date = config["general"]["current_date"]
    else:
        current_date = Date(config["general"]["current_date"], calendar)
    delta_date = (
        config["general"]["nyear"],
        config["general"]["nmonth"],
        config["general"]["nday"],
        config["general"]["nhour"],
        config["general"]["nminute"],
        config["general"]["nsecond"],
    )

    config["general"]["delta_date"] = delta_date
    config["general"]["current_date"] = current_date
    config["general"]["start_date"] = current_date
    config["general"]["initial_date"] = Date(
        config["general"]["initial_date"], calendar
    )
    config["general"]["final_date"] = Date(config["general"]["final_date"], calendar)
    config["general"]["prev_date"] = current_date - (0, 0, 1, 0, 0, 0)

    config["general"]["next_date"] = current_date.add(delta_date)
    config["general"]["last_start_date"] = current_date - delta_date
    config["general"]["end_date"] = config["general"]["next_date"] - (0, 0, 1, 0, 0, 0)

    config["general"]["runtime"] = (
        config["general"]["next_date"] - config["general"]["current_date"]
    )

    config["general"]["total_runtime"] = (
        config["general"]["next_date"] - config["general"]["initial_date"]
    )

    config["general"]["run_datestamp"] = (
        config["general"]["current_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        + "-"
        + config["general"]["end_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
    )

    config["general"]["last_run_datestamp"] = (
        config["general"]["last_start_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        + "-"
        + config["general"]["prev_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
    )
    return config


def _add_all_folders(config):
    all_filetypes = [
        "analysis",
        "config",
        "log",
        "mon",
        "couple",
        "scripts",
        "ignore",
        "unknown",
        "src",
    ]
    config["general"]["out_filetypes"] = [
        "analysis",
        "log",
        "mon",
        "scripts",
        "ignore",
        "unknown",
        "outdata",
        "restart_out",
    ]
    config["general"]["in_filetypes"] = [
        "scripts",
        "input",
        "forcing",
        "bin",
        "config",
        "restart_in",
    ]
    config["general"]["reusable_filetypes"] = config["general"].get(
        "reusable_filetypes", ["bin", "src"]
    )
    # Define the files that could be reusable accross runs (external files)
    config["general"]["potentially_reusable_filetypes"] = (
        all_filetypes + config["general"]["in_filetypes"]
    )
    # Apply changes from ``--update-files`` flag
    config = helpers.update_reusable_filetypes(config)

    config["general"]["thisrun_dir"] = (
        config["general"]["experiment_dir"]
        + "/run_"
        + config["general"]["run_datestamp"]
    )

    for filetype in all_filetypes:
        config["general"]["experiment_" + filetype + "_dir"] = (
            config["general"]["experiment_dir"] + "/" + filetype + "/"
        )

    all_filetypes.append("work")
    config["general"]["all_filetypes"] = all_filetypes

    for filetype in all_filetypes:
        config["general"]["thisrun_" + filetype + "_dir"] = (
            config["general"]["thisrun_dir"] + "/" + filetype + "/"
        )

    config["general"]["work_dir"] = config["general"]["thisrun_work_dir"]

    all_model_filetypes = [
        "analysis",
        "bin",
        "config",
        "forcing",
        "input",
        "couple",
        "log",
        "mon",
        "outdata",
        "restart_in",
        "restart_out",
        "viz",
        "ignore",
    ]

    config["general"]["all_model_filetypes"] = all_model_filetypes

    for model in config["general"]["valid_model_names"]:
        for filetype in all_model_filetypes:
            if "restart" in filetype:
                filedir = "restart"
            else:
                filedir = filetype
            config[model]["experiment_" + filetype + "_dir"] = (
                config["general"]["experiment_dir"] + "/" + filedir + "/" + model + "/"
            )
            config[model]["thisrun_" + filetype + "_dir"] = (
                config["general"]["thisrun_dir"] + "/" + filedir + "/" + model + "/"
            )
            config[model]["all_filetypes"] = all_model_filetypes

    return config


def set_prev_date(config):
    import esm_parser

    """Sets several variables relevant for the previous date. Loops over all models in ``valid_model_names``, and sets model variables for:
    * ``prev_date``
    """
    for model in config["general"]["valid_model_names"]:
        if "time_step" in config[model] and not (
            isinstance(config[model]["time_step"], str)
            and "${" in config[model]["time_step"]
        ):
            config[model]["prev_date"] = config["general"]["current_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(config[model]["time_step"]),
            )
            config[model]["last_parent_date"] = config[model]["prev_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(config[model]["time_step"]),
            )

        # NOTE(PG, MAM): Here we check if the time step still has a variable which might be set in a different model, and resolve this case
        elif "time_step" in config[model] and (
            isinstance(config[model]["time_step"], str)
            and "${" in config[model]["time_step"]
        ):
            dt = esm_parser.find_variable(
                model, config[model]["time_step"], config, [], []
            )
            config[model]["prev_date"] = config["general"]["current_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(dt),
            )
            config[model]["last_parent_date"] = config[model]["prev_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(dt),
            )

        else:
            config[model]["prev_date"] = config["general"]["current_date"]
    return config


def set_parent_info(config):

    """Sets several variables relevant for the previous date. Loops over all models in ``valid_model_names``, and sets model variables for:
    * ``parent_expid``
    * ``parent_date``
    * ``parent_restart_dir``
    """

    # Make sure "ini_parent_dir" and "ini_restart_dir" both work:
    for model in config["general"]["valid_model_names"]:
        if not "ini_parent_dir" in config[model]:
            if "ini_restart_dir" in config[model]:
                config[model]["ini_parent_dir"] = config[model]["ini_restart_dir"]
        if not "ini_parent_exp_id" in config[model]:
            if "ini_restart_exp_id" in config[model]:
                config[model]["ini_parent_exp_id"] = config[model]["ini_restart_exp_id"]
        if not "ini_parent_date" in config[model]:
            if "ini_restart_date" in config[model]:
                config[model]["ini_parent_date"] = config[model]["ini_restart_date"]

    # check if parent is defined in esm_tools style
    # (only given for setup)
    setup = config["general"]["setup_name"]
    if not setup in config:
        setup = "general"
    if "ini_parent_exp_id" in config[setup]:
        for model in config["general"]["valid_model_names"]:
            if not "ini_parent_exp_id" in config[model]:
                config[model]["ini_parent_exp_id"] = config[setup]["ini_parent_exp_id"]
    if "ini_parent_date" in config[setup]:
        for model in config["general"]["valid_model_names"]:
            if not "ini_parent_date" in config[model]:
                config[model]["ini_parent_date"] = config[setup]["ini_parent_date"]
    if "ini_parent_dir" in config[setup]:
        for model in config["general"]["valid_model_names"]:
            if not "ini_parent_dir" in config[model]:
                config[model]["ini_parent_dir"] = (
                    config[setup]["ini_parent_dir"] + "/" + model
                )

    # Get correct parent info
    for model in config["general"]["valid_model_names"]:
        if config[model]["lresume"] == True and config["general"]["run_number"] == 1:
            config[model]["parent_expid"] = config[model]["ini_parent_exp_id"]
            if "parent_date" not in config[model]:
                config[model]["parent_date"] = config[model]["ini_parent_date"]
            config[model]["parent_restart_dir"] = config[model]["ini_parent_dir"]
        else:
            config[model]["parent_expid"] = config["general"]["expid"]
            if "parent_date" not in config[model]:
                config[model]["parent_date"] = config[model]["prev_date"]
            config[model]["parent_restart_dir"] = config[model][
                "experiment_restart_in_dir"
            ]
    return config


def finalize_config(config):
    config.finalize()
    return config


def add_submission_info(config):

    bs = batch_system.batch_system(config, config["computer"]["batch_system"])

    submitted = bs.check_if_submitted()
    if submitted:
        jobid = bs.get_jobid()
    else:
        jobid = os.getpid()

    config["general"]["submitted"] = submitted
    config["general"]["jobid"] = jobid
    return config


def initialize_batch_system(config):

    config["general"]["batch"] = batch_system.batch_system(
        config, config["computer"]["batch_system"]
    )

    return config


def initialize_coupler(config):
    if config["general"]["standalone"] == False:
        from . import coupler

        for model in list(config):
            if model in coupler.known_couplers:
                config["general"]["coupler_config_dir"] = (
                    f"{config['general']['base_dir']}"
                    f"/{config['general']['expid']}"
                    f"/run_{config['general']['run_datestamp']}"
                    f"/config/{model}/"
                )
                config["general"]["coupler"] = coupler.coupler_class(config, model)
                break
        config["general"]["coupler"].add_files(config)
    return config


def set_logfile(config):
    log_file_path = (
        f"{config['general']['experiment_log_dir']}"
        f"/{config['general']['expid']}"
        f"_{config['general']['setup_name']}.log"
    )
    config["general"]["experiment_log_file"] = config["general"].get(
        "experiment_log_file", log_file_path
    )
    return config



import copy
import difflib
import glob
import os
import re
import shutil
import time
import colorama
import yaml
from esm_parser import determine_computer_from_hostname
from esm_runscripts import color_diff
from loguru import logger

from .read_shipped_data import *
from .test_utilities import *

# Bold strings
bs = "\033[1m"
be = "\033[0m"

# Define default files for comparisson
compare_files = {"comp": ["comp-"], "run": [".run", "finished_config", "namelists"]}

"""
``config.yaml``:
    This file includes exceptions rules to be applied during the tests (such as copying
    instead of downloading the source code for compilation), and output and restart
    files to be checked that they exist after an actual run. This file needs to exist
    in each ``resources/runscripts/<model>`` folder.
"""


class Comparison:
    """Compares two ESM Tools files"""

    def __init__(self, test_file, truth_file):
        """
        Parameters
        ----------
        test_file : str
            str representation (already opened and read in) of the file (e.g.
            run file, compilatoin script, namelist) which you want to check
        truth_file : str
            str representation of the file which is known to be a valid truth.
        """
        self.test_file = test_file
        self.truth_file = truth_file

    @classmethod
    def from_filepaths(cls, test_file, truth_file):
        with open(test_file, "r") as f1:
            test_file = f1.read()
        with open(test_file, "r") as f2:
            truth_file = f2.read()
        return cls(test_file, truth_file)

    @classmethod
    def from_test_filepath_and_pkg(cls, test_file, truth_file):
        """
        Given a test filepath, and a relative truth path, return a new Comparison.

        For the relative truth path, we are reading from the package. So,
        assuming you want to use the following as your truth:

        >>> truth_file = "ollie/run/awicm/awicm2-initial-monthly/scripts/awicm2-initial-monthly_compute_20000101 -20000131.run"

        This would check the run file for a run of awicm using a awicm2
        initialization with monthly restarts which is run on the ollie HPC.
        """
        # get last tested (Truth)
        with open(test_file, "r") as f:
            test_file = f1.read()
        truth_file = read_shipped_data.get_last_tested(truth_file)
        return cls(test_file, truth_file)


class CompileFileComparison(Comparison):
    pass


class SadFileComparison(Comparison):
    pass


class FinishedESMConfigComparison(Comparison):
    pass


def NamelistsComparison(Comparison):
    pass


#######################################################################################
# TESTS
#######################################################################################
def comp_test(info):
    """
    Downloads/copies the source codes of the models to be tested and compiles them. If
    the ``--check`` flag is provided in the ``esm_tests`` call, it does not
    download/copy and compile the source codes, but produces the ``comp-*.sh``
    compilation scripts instead, for comparison purposes with last tested versions of
    them.

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info.
    """
    # Load dictionaries from ``info``
    scripts_info = info["scripts"]
    user_info = info["user"]
    this_computer = info["this_computer"]

    # Set the regex format for check compilations
    cd_format = re.compile("         cd (.*)")

    # Set the counter to 0
    c = 0
    # Loop through the models and their scripts
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        # Loop through the scripts of each model
        for script, v in scripts.items():
            # If the user has specified the ``hold`` flag, wait for input
            if info["hold"] and c != 0:
                input("Press ENTER to continue...")
            # Calculate progress
            c += 1
            progress = round(c / scripts_info["general"]["num_scripts"] * 100, 1)

            # Define variables from the runscript
            version = v["version"]
            if v["comp_command"]:
                comp_command = f"{v['comp_command']} --no-motd -k"
            else:
                comp_command = f"esm_master comp-{model}-{version} --no-motd -k"
            general_model_dir = f"{user_info['test_dir']}/comp/{model}"
            model_dir = f"{user_info['test_dir']}/comp/{model}/{model}-{version}"
            logger.info(f"\tCOMPILING ({progress}%) {model}-{version}:")

            # Create the folders where this test should run and change directory
            if not os.path.isdir(general_model_dir):
                os.makedirs(general_model_dir)
            os.chdir(general_model_dir)

            # Check if the test compilation has been already run (the compilation
            # directory already exists). This is particularly important for different
            # testing scripts that use the same version of a model, so that compilation
            # is not run more than once
            if os.path.isdir(model_dir):
                v["action"] = {"comp": "Directory already exists"}
                logger.info(f"\t\tDirectory already exists, skipping")
                with open(f"{model_dir}/comp.out") as o:
                    out = o.read()
            else:
                # ACTUAL COMPILATION
                # ==================
                # Get the source code if actual compilation is required
                if info["actually_compile"]:
                    # Downloading or copying
                    # ----------------------
                    # Load the ``config.yaml`` for this model to find if the source
                    # code needs to be downloaded or copied
                    with open(f"{os.path.dirname(v['path'])}/config.yaml", "r") as cf:
                        config_test = yaml.load(cf, Loader=yaml.FullLoader)
                    copying = (
                        config_test["comp"]
                        .get("cp_instead_of_download", {})
                        .get(this_computer, {})
                        .get(f"{model}-{version}")
                    )
                    # If copying of the source code is required go ahead, copy and,
                    # if necessary, untar
                    if copying:
                        logger.info("\t\tCopying")
                        out = ""
                        if os.path.isdir(copying):
                            shutil.copytree(copying, general_model_dir)
                        else:
                            shutil.copy2(copying, general_model_dir)
                        if copying.endswith("tar.gz"):
                            logger.info("\t\tUntaring")
                            file2untar = (
                                f"{general_model_dir}/{os.path.basename(copying)}"
                            )
                            out += sh(f"tar -xvf {file2untar}")
                            os.remove(file2untar)
                    # If download of the source code is required, run
                    # ``esm_master get-<model>-<version>``
                    else:
                        get_command = f"esm_master get-{model}-{version} --no-motd"
                        logger.info("\t\tDownloading")
                        out = sh(get_command)
                        if "Traceback (most recent call last):" in out:
                            logger.error(f"\t\t\tProblem downloading!\n\n{out}")

                # CHECK COMPILATION
                # =================
                # No actual compilation: for a check compilation where no real
                # compilation occurs, but the ``comp-`` scripts are generated, we need
                # to trick esm_master into thinking that all the folders source code
                # folders are available, even if they were not downloaded or copied
                else:
                    # Evaluate and create folders to trick esm_master
                    # -----------------------------------------------
                    # Run a ``esm_master --check``
                    out = sh(f"{comp_command} -c")
                    folders = []
                    # Search for the folders to be created
                    for line in out.split("\n"):
                        if "cd" in line and "cd .." not in line:
                            found_format = cd_format.findall(line)
                            if len(found_format) > 0:
                                if (
                                    ";" not in found_format[0]
                                    and "/" not in found_format[0]
                                ):
                                    folders.append(found_format[0])
                    if len(folders) == 0:
                        logger.warning(
                            f'NOT TESTING {model + version}: "cd" command not found'
                        )
                        continue

                    # Make the folders needed to trick esm_master
                    prim_f = folders[0]
                    folders.append(f"{model}-{version}")
                    folders = [x for x in set(folders)]
                    os.mkdir(prim_f)
                    for folder in folders:
                        os.mkdir(prim_f + "/" + folder)

                    # Get files from the
                    # ``esm_test/runscripts/<model>/comp_files4check_runs`` (i.e.
                    # namelists that are hosted in another repository could be placed
                    # there so that checks can run successfully without having to
                    # download the code)
                    copy_comp_files4check_runs(
                        script, v, f"{general_model_dir}/{prim_f}"
                    )

                # Compile
                if info["actually_compile"]:
                    logger.info("\t\tCompiling")
                else:
                    logger.info("\t\tWritting compilation scripts")
                out = sh(comp_command)

                # Write output file
                with open(f"{model_dir}/comp.out", "w") as o:
                    o.write(out)

                # Move and cleanup files
                for f in os.listdir(general_model_dir):
                    if "comp-" in f and not os.path.isfile(f"{model_dir}/{f}"):
                        shutil.move(f"{general_model_dir}/{f}", model_dir)
                    elif "comp-" in f and os.path.isfile(f"{model_dir}/{f}"):
                        os.remove(f"{general_model_dir}/{f}")
                    if f == "dummy_script.sh":
                        os.remove(f"{general_model_dir}/{f}")

            # Checks
            success = check(info, "comp", model, version, out, script, v)
            if success:
                logger.info("\t\tSuccess!")


def run_test(info):
    """
    Runs the simulations and notifies when they are finished. If the ``--check`` flag
    is given in the ``esm_tests`` call, check simulations are run
    (``esm_runscripts -c``).

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info.
    """
    # Load dictionaries from ``info``
    scripts_info = info["scripts"]
    user_info = info["user"]
    actually_run = info["actually_run"]

    # Set the counter to 0
    c = 0
    submitted = []
    # Loop through the models and their scripts
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        # Loop through the scripts of each model
        for script, v in scripts.items():
            # If the user has specified the ``hold`` flag, wait for input
            if info["hold"] and c != 0:
                input("Press ENTER to continue...")
            c += 1
            progress = round(c / scripts_info["general"]["num_scripts"] * 100, 1)

            # Define variables from the runscript
            version = v["version"]
            runscript_path = v["path"]
            general_run_dir = f"{user_info['test_dir']}/run/{model}/"
            run_dir = f"{general_run_dir}/{script}"
            model_dir = f"{user_info['test_dir']}/comp/{model}/{model}-{version}"
            logger.info(f"\tSUBMITTING ({progress}%) {model}/{script}:")

            # Create the folders where this test should run and change directory
            if not os.path.isdir(general_run_dir):
                os.makedirs(general_run_dir)

            # If the simulation already exists, only read the ``.out`` file
            if os.path.isdir(run_dir):
                v["action"]["submission"] = "Directory already exists"
                logger.info(f"\t\tDirectory already exists, skipping")
                with open(f"{run_dir}/run.out", "r") as o:
                    out = o.read()
                if actually_run:
                    submitted.append((model, script))
            # If the simulation does not exists, run it or check it
            else:
                # Change the path to that of the runscript
                os.chdir(os.path.dirname(runscript_path))

                # Define the check flag if necessary
                additional_actions = ""
                if actually_run:
                    check_flag = ""
                else:
                    check_flag = "-c"
                    # In case of iterative coupling, submit the subscripts too
                    if v["iterative_coupling"]:
                        for imodel, mdata in v["iterative_models"].items():
                            additional_actions = (
                                f"{additional_actions}; "
                                f"echo {'='*10}; echo {imodel}; echo {'='*10}; "
                                f"esm_runscripts "
                                f"{os.path.dirname(v['path'])}/{mdata['script']} "
                                f"-e {script} --open-run {check_flag}"
                            )
                # Export test variables and run the simulation
                env_vars = [
                    f"ACCOUNT='{user_info['account']}'",
                    f"ESM_TESTING_DIR='{general_run_dir}'",
                    f"MODEL_DIR='{model_dir}'",
                ]
                run_command = (
                    f"esm_runscripts {v['path']} -e {script} --open-run "
                    f"{check_flag}{additional_actions}"
                )
                out = sh(run_command, env_vars)

                # Write output file
                if os.path.isdir(run_dir):
                    with open(f"{run_dir}/run.out", "w") as o:
                        o.write(out)

                if actually_run:
                    submitted.append((model, script))

            # Check submission
            success = check(info, "submission", model, version, out, script, v)

    # Check if simulations are finished
    total_sub = len(submitted)
    subc = 1
    if total_sub > 0:
        logger.info(
            "\nWaiting for submitted runs to finish... You can choose to cancel the "
            + "script now and come back to it at a later state with the same command "
            + "you submitted it, but remember to remove the '-d' so that nothing is "
            + "deleted and this script can be resumed from where it was."
        )
    infoc = 10
    while submitted:
        cc = 0
        finished_runs = []
        # Loop through models and scripts
        for model, script in submitted:
            v = scripts_info[model][script]
            progress = round(subc / total_sub * 100, 1)
            exp_dir = f"{user_info['test_dir']}/run/{model}/{script}/"
            exp_dir_log = f"{exp_dir}/log/"
            nmodels_success = 1
            # Search through the ``_compute_`` files for a string that indicates that
            # the run has finished
            for f in os.listdir(exp_dir_log):
                if "_compute_" in f and ".log" in f:
                    with open(f"{exp_dir_log}/{f}") as m:
                        observe_out = m.read()
                        # If the run has finished successfully label the state for
                        # ``run_finished`` as ``True`` and run a check for files that
                        # should have been created
                        if "Reached the end of the simulation, quitting" in observe_out:
                            if nmodels_success == v["nmodels_iterative_coupling"]:
                                logger.info(
                                    f"\tRUN FINISHED ({progress}%) {model}/{script}"
                                )
                                logger.info(f"\t\tSuccess!")
                                finished_runs.append(cc)
                                subc += 1
                                v["state"]["run_finished"] = True
                                # Run a check for run finished
                                success = check(
                                    info,
                                    "run",
                                    model,
                                    version,
                                    "",
                                    script,
                                    v,
                                )
                            else:
                                nmodels_success += 1
                        # If the run has errors label the state for ``run_finished`` as
                        # ``False`` and run a check for files that should have been
                        # created anyway
                        elif "ERROR:" in observe_out:
                            logger.info(
                                f"\tRUN FINISHED ({progress}%) {model}/{script}"
                            )
                            logger.error(f"\t\tSimulation crashed!")
                            finished_runs.append(cc)
                            subc += 1
                            v["state"]["run_finished"] = False
                            # Run a check for run finished
                            success = check(
                                info,
                                "run",
                                model,
                                version,
                                "",
                                script,
                                v,
                            )
            # if not info["keep_run_folders"]:
            #    folders_to_remove = [
            #        "run_",
            #        "restart",
            #        "outdata",
            #        "input",
            #        "forcing",
            #        "unknown",
            #    ]
            #    logger.debug(f"\t\tDeleting {folders_to_remove}")
            #    for folder in os.listdir(exp_dir):
            #        for fr in folders_to_remove:
            #            if fr in folder:
            #                shutil.rmtree(f"{exp_dir}/{folder}")
            #                continue

            # Update the testing index for the next iteration
            cc += 1

        # Remove finished runs from the ``submitted`` list
        finished_runs = list(set(finished_runs))
        for indx in finished_runs[::-1]:
            del submitted[indx]

        # Wait 10 iterations to notify the state
        if infoc == 10 and len(submitted) > 0:
            infoc = 0
            runs = ""
            for model, script in submitted:
                runs += f"\t- {model}/{script}\n"
            logger.info(f"\nWaiting for the following runs to finish:\n{runs}")
        else:
            infoc += 1

        # Wait 30 seconds to check again the state of the submitted tests
        if len(submitted) > 0:
            time.sleep(30)


#######################################################################################
# CHECKS
#######################################################################################
def check(info, mode, model, version, out, script, v):
    """
    Checks for errors, missing files and compares current files with previous files in
    ``resources/last_tested``. Behaves differently for different modes.

    Modes:
        * ``comp``: checks compilation scripts
        * ``submission``: checks submission files (``.run``, namelists, ...)
        * ``run``: same as ``submission`` but also checks for files produced by the
            simulation (restarts and output)

    Parameters
    ----------
    info : dict
        Dictionary that contains the testing info.
    mode : str
        Mode of the check.
    model : str
        Model for which the check is run.
    version : str
        Version of the model for which the check is run.
    out : str
        Output of the shell operation that is to be checked.
    script : str
        Name of the script to be checked.
    v : dict
        Dictionary containing script variables such as ``version``, ``comp_command``, ...

    Returns
    -------
    success : bool
        ``True`` if all checks were passed, ``False`` if not.
    """
    # Set variables
    success = True
    mode_name = {"comp": "compilation", "submission": "submission", "run": "runtime"}
    last_tested_dir = info["last_tested_dir"]
    this_computer = info["this_computer"]
    user_info = info["user"]
    actually_run = info["actually_run"]

    # Load config for this mode
    with open(f"{os.path.dirname(v['path'])}/config.yaml", "r") as c:
        config_test = yaml.load(c, Loader=yaml.FullLoader)

    # There are 3 modes: comp, submission and run, and 2 config_modes: comp and run.
    # This is because submission shares a lot of operations with run so they are in the
    # same config_mode group
    if mode == "submission":
        config_mode = "run"
    else:
        config_mode = mode

    # Check if the config_mode is defined in the ``config.yaml`` file, and load the
    # info for this type of test
    if config_mode not in config_test:
        logger.error(
            f"Missing '{mode}' section in '{os.path.dirname(v['path'])}/config.yaml'!"
        )
    config_test = config_test[config_mode]

    # Set mode variables
    if mode == "comp":
        if info["actually_compile"]:
            test_type = "actual"
        else:
            test_type = "check"
        actually_do = info["actually_compile"]
        subfolder = f"{model}-{version}"
    elif mode == "run":
        if actually_run:
            test_type = "actual"
        else:
            test_type = "check"
        actually_do = actually_run
        subfolder = script
    elif mode == "submission":
        # Do not perform the file checks before the simulation is finished
        if actually_run:
            test_type = "actual"
        else:
            test_type = "check"
        actually_do = False
        subfolder = script

    # Check for errors
    if actually_do or mode == "submission":
        # Check for errors in the output
        errors = config_test.get(test_type, {}).get("errors", [])
        # Add specific errors
        if mode == "comp":
            errors.append("errors occurred!")
        if mode == "submission":
            errors.extend(["Traceback (most recent call last):", "ERROR"])
        # Loop through errors
        for error in errors:
            if error in out:
                logger.error(f"\t\tError during {mode_name[mode]}!\n\n{out}")
                success = False
        if mode != "run":
            v["state"][mode] = success
    else:
        if mode == "comp":
            # No need to check for this in check mode
            v["state"][mode] = success

    # Check for missing files during an actual operation (not a check, not submission)
    if actually_do:
        # Check if files exist
        files_checked = exist_files(
            config_test.get(test_type, {}).get("files", []),
            f"{user_info['test_dir']}/{mode}/{model}/{subfolder}",
        )
        v["state"][f"{mode}_files"] = files_checked
        success = success and files_checked

    # Compare scripts with previous, if existing
    this_compare_files = copy.deepcopy(compare_files[config_mode])
    # TODO: The iterative coupling needs a rework. Therefore, no testing for files
    # is develop. Include the tests after iterative coupling is reworked
    if config_mode == "run" and v["iterative_coupling"]:
        this_compare_files = []
    this_compare_files.extend(config_test.get(test_type, {}).get("compare", []))
    this_test_dir = f"{config_mode}/{model}/{subfolder}/"
    v["state"][f"{mode}_files_identical"] = True
    # Loop through the files to be compared
    for cfile in this_compare_files:
        # Load lines to be ignored
        ignore_lines = info["ignore"].get(cfile, [])
        # Get relative paths of the files to be compared
        subpaths_source, subpaths_target = get_rel_paths_compare_files(
            info, cfile, v, this_test_dir
        )
        for sp, sp_t in zip(subpaths_source, subpaths_target):
            # Check if the source exists (this simulation file)
            if not os.path.isfile(f"{user_info['test_dir']}/{sp}"):
                logger.error(f"\t\t'{sp}' file is missing!")
                identical = False
            else:
                # Check if the target exists (in last_tested folder)
                if os.path.isfile(f"{last_tested_dir}/{this_computer}/{sp_t}"):
                    # Check if files are identical
                    identical, differences = print_diff(
                        f"{last_tested_dir}/{this_computer}/{sp_t}",
                        f"{user_info['test_dir']}/{sp}",
                        sp,
                        ignore_lines,
                    )
                    success += identical
                    # Update state dictionaries
                    if not identical:
                        v["differences"] = v.get("differences", {})
                        v["differences"][config_mode] = v["differences"].get(
                            config_mode, {}
                        )
                        v["differences"][config_mode][sp] = differences
                        v["state"][f"{mode}_files_identical"] &= False
                    else:
                        v["state"][f"{mode}_files_identical"] &= True
                else:
                    logger.warning(f"\t\t'{sp_t}' file not yet in 'last_tested'")

    return success


def exist_files(files, path):
    files_checked = True
    for f in files:
        if "*" in f:
            listing = glob.glob(f"{path}/{f}")
            if len(listing) == 0:
                logger.error(f"\t\tNo files following the pattern '{f}' were created!")
                files_checked = False
        else:
            if not os.path.isfile(f"{path}/{f}"):
                logger.error(f"\t\t'{f}' does not exist!")
                files_checked = False
    return files_checked


def get_rel_paths_compare_files(info, cfile, v, this_test_dir):
    # Load relevant variables from ``info``
    user_info = info["user"]
    # Initialize ``subpaths`` list
    subpaths = []
    # If the file type is ``comp-``, append all the files that match that string to
    # ``subpaths``
    if cfile == "comp-":
        for f in os.listdir(f"{user_info['test_dir']}/{this_test_dir}"):
            if cfile in f:
                subpaths.append(f"{this_test_dir}/{f}")
        if len(subpaths) == 0:
            logger.error("\t\tNo 'comp-*.sh' file found!")
    # If the file type matches ``.run`` or ``finished_config``, append all the files
    # that match that string to ``subpaths``. These files are taken always from the
    # first ``run_`` folder so that a check test and an actual test files are
    # comparable
    elif cfile in [".run", "finished_config"]:
        files_to_folders = {".run": "scripts", "finished_config": "config"}
        ctype = files_to_folders[cfile]
        ldir = os.listdir(f"{user_info['test_dir']}/{this_test_dir}")
        ldir.sort()
        for f in ldir:
            # Take the first run directory
            if "run_" in f:
                cf_path = f"{this_test_dir}/{f}/{ctype}/"
                cfiles = glob.glob(f"{user_info['test_dir']}/{cf_path}/*{cfile}*")
                # If not found, try in the general directory
                if len(cfiles) == 0:
                    cf_path = f"{this_test_dir}/{ctype}/"
                    cfiles = glob.glob(f"{user_info['test_dir']}/{cf_path}/*{cfile}*")
                # Make sure we always take the first run
                cfiles.sort()
                num = 0
                if os.path.islink(
                    f"{user_info['test_dir']}/{cf_path}/{cfiles[num].split('/')[-1]}"
                ):
                    num = 1
                subpaths.append(f"{cf_path}/{cfiles[num].split('/')[-1]}")
                break
    elif cfile == "namelists":
        # Get path of the finished_config
        s_config_yaml, _ = get_rel_paths_compare_files(
            info, "finished_config", v, this_test_dir
        )
        namelists = extract_namelists(f"{user_info['test_dir']}/{s_config_yaml[0]}")
        ldir = os.listdir(f"{user_info['test_dir']}/{this_test_dir}")
        ldir.sort()
        for f in ldir:
            # Take the first run directory
            if "run_" in f:
                cf_path = f"{this_test_dir}/{f}/work/"
                for n in namelists:
                    if not os.path.isfile(f"{user_info['test_dir']}/{cf_path}/{n}"):
                        logger.debug(f"'{cf_path}/{n}' does not exist!")
                    subpaths.append(f"{cf_path}/{n}")
                break
    else:
        subpaths = [f"{this_test_dir}/{cfile}"]

    # Remove run directory from the targets
    subpaths_source = subpaths
    subpaths_target = []
    datestamp_format = re.compile(r"_[\d]{8}-[\d]{8}$")
    for sp in subpaths:
        sp_t = ""
        pieces = sp.split("/")
        for p in pieces:
            if "run_" not in p:
                sp_t += f"/{p}"
        # Remove the datestamp
        if datestamp_format.findall(sp_t):
            sp_t = sp_t.replace(datestamp_format.findall(sp_t)[0], "")
        subpaths_target.append(sp_t)

    return subpaths_source, subpaths_target


def extract_namelists(s_config_yaml):
    # Read config file
    with open(s_config_yaml, "r") as c:
        config = yaml.load(c, Loader=yaml.FullLoader)

    namelists = []
    for component in config.keys():
        namelists_component = config[component].get("namelists", [])
        for nml in namelists_component:
            config_sources = config[component].get("config_sources", {})
            if (
                nml in config_sources
                or any([nml in x for x in config_sources.values()])
            ):
                namelists.append(nml)

    return namelists


#######################################################################################
# OUTPUT
#######################################################################################
def print_diff(sscript, tscript, name, ignore_lines):
    script_s = open(sscript).readlines()
    script_t = open(tscript).readlines()

    # Check for ignored lines
    new_script_s = []
    for line in script_s:
        ignore_this = False
        for iline in ignore_lines:
            if iline in line:
                ignore_this = True
        if not ignore_this:
            new_script_s.append(line)
    script_s = new_script_s
    new_script_t = []
    for line in script_t:
        ignore_this = False
        for iline in ignore_lines:
            if iline in line:
                ignore_this = True
        if not ignore_this:
            new_script_t.append(line)
    script_t = new_script_t

    diffobj = difflib.SequenceMatcher(a=script_s, b=script_t)
    differences = ""
    if diffobj.ratio() == 1:
        logger.info(f"\t\t'{name}' files are identical")
        identical = True
    else:
        # Find differences
        pdifferences = ""
        for line in color_diff(difflib.unified_diff(script_s, script_t)):
            differences += line
            pdifferences += f"\t\t{line}"

        logger.info(f"\n\tDifferences in {name}:\n{pdifferences}\n")
        # input("Press enter to continue...")
        identical = False

    return identical, differences


def save_files(info, user_choice):
    # Load relevant variables from ``info``
    scripts_info = info["scripts"]
    actually_run = info["actually_run"]
    user_info = info["user"]
    last_tested_dir = info["last_tested_dir"]
    runscripts_dir = get_runscripts_dir()
    this_computer = info["this_computer"]
    # ``user_choice`` is set through the ``--save`` flag. If the flag is not used the
    # main program reaches this point and asks the user about saving files in
    # ``last_tested``. If the flag is used alone or the value is set to ``True``, then
    # it does save without asking. If the flag is used and the value is set to ``False``
    # then it doesn't save and it doesn't ask
    if not user_choice:
        not_answered = True
        while not_answered:
            # Ask if saving of the compared files is required
            save = input(
                f"Would you like to save the files in the '{last_tested_dir}' folder for later "
                + "comparisson and/or committing to GitHub?[y/n]: "
            )
            if save == "y":
                not_answered = False
            elif save == "n":
                logger.info(f"No files will be saved in '{last_tested_dir}'")
                not_answered = False
                return
            else:
                print(f"'{save}' is not a valid answer!")

    # Select test types
    if info["actually_compile"]:
        test_type_c = "actual"
    else:
        test_type_c = "check"
    if actually_run:
        test_type_r = "actual"
    else:
        test_type_r = "check"

    logger.info(f"Saving files to '{last_tested_dir}'...")
    # Loop through models
    for model, scripts in scripts_info.items():
        # Ignore ``general``
        if model == "general":
            continue
        # Load the model's ``config.yaml``
        model_config = f"{runscripts_dir}/{model}/config.yaml"
        if not os.path.isfile(model_config):
            logger.error(f"'{model_config}' not found!")
        with open(model_config, "r") as c:
            config_test = yaml.load(c, Loader=yaml.FullLoader)
        # Get name patters from the files to be compared
        compare_files_comp = copy.deepcopy(compare_files["comp"])
        compare_files_comp.extend(
            config_test.get("comp", {}).get(test_type_c, {}).get("compare", [])
        )
        compare_files_run = copy.deepcopy(compare_files["run"])
        # TODO: The iterative coupling needs a rework. Therefore, no testing for files
        # is develop. Include the tests after iterative coupling is reworked
        if next(iter(scripts.values()))["iterative_coupling"]:
            compare_files_run = []
        compare_files_run.extend(
            config_test.get("run", {}).get(test_type_r, {}).get("compare", [])
        )
        # Loop through scripts
        for script, v in scripts.items():
            version = v["version"]
            # Loop through comp and run
            for mode in ["comp", "run"]:
                if mode == "comp":
                    this_compare_files = compare_files_comp
                    subfolder = f"{model}-{version}"
                elif mode == "run":
                    this_compare_files = compare_files_run
                    subfolder = f"{script}"
                this_test_dir = f"{mode}/{model}/{subfolder}/"
                # Loop through comparefiles
                for cfile in this_compare_files:
                    subpaths_source, subpaths_target = get_rel_paths_compare_files(
                        info, cfile, v, this_test_dir
                    )
                    for sp, sp_t in zip(subpaths_source, subpaths_target):
                        if os.path.isfile(f"{last_tested_dir}/{sp_t}"):
                            logger.debug(
                                f"\t'{sp_t}' file in '{last_tested_dir}' will be overwritten"
                            )
                        if not os.path.isdir(
                            os.path.dirname(f"{last_tested_dir}/{this_computer}/{sp_t}")
                        ):
                            os.makedirs(
                                os.path.dirname(
                                    f"{last_tested_dir}/{this_computer}/{sp_t}"
                                )
                            )
                        shutil.copy2(
                            f"{user_info['test_dir']}/{sp}",
                            f"{last_tested_dir}/{this_computer}/{sp_t}",
                        )
    # Load current state
    with open(get_state_yaml_path(), "r") as st:
        current_state = yaml.load(st, Loader=yaml.FullLoader)
    # Update with this results
    results = format_results(info)
    current_state = deep_update(current_state, results)
    current_state = sort_dict(current_state)
    with open(get_state_yaml_path(), "w") as st:
        state = yaml.dump(current_state)
        st.write(state)


def print_results(results, info):
    colorama.init(autoreset=True)

    if info.get("bulletpoints", False):
        bp = "- "
    else:
        bp = ""

    logger.info("")
    logger.info("")
    logger.info(f"{bs}RESULTS{be}")
    logger.info("")
    for model, versions in results.items():
        logger.info(f"{bp}{colorama.Fore.CYAN}{model}:")
        for version, scripts in versions.items():
            logger.info(f"    {bp}{colorama.Fore.MAGENTA}{version}:")
            for script, computers in scripts.items():
                logger.info(f"        {bp}{colorama.Fore.WHITE}{script}:")
                for computer, data in computers.items():
                    if data["compilation"]:
                        compilation = f"{colorama.Fore.GREEN}compiles"
                    else:
                        compilation = f"{colorama.Fore.RED}compilation failed"
                    if data["run"]:
                        run = f"{colorama.Fore.GREEN}runs"
                    else:
                        run = f"{colorama.Fore.RED}run failed"
                    logger.info(
                        f"            {bp}{colorama.Fore.WHITE}{computer}:\t{compilation}\t{run}"
                    )
    logger.info(f"{colorama.Fore.WHITE}")
    logger.info("")


def format_results(info):
    scripts_info = info["scripts"]
    results = {}
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        results[model] = {}
        for script, v in scripts.items():
            version = v["version"]
            results[model][version] = results[model].get(version, {})
            results[model][version][script] = results[model][version].get(script, {})
            state = v["state"]
            compilation = (
                state["comp"]
                and state.get("comp_files", True)
                and state.get("comp_files_identical", True)
            )
            run = (
                state.get("run_finished", True)
                and state.get("run_files", True)
                and state["submission"]
                and state.get("submission_files_identical", True)
            )
            results[model][version][script][info["this_computer"]] = {
                "compilation": compilation,
                "run": run,
            }

    return results


def sort_dict(dict_to_sort):
    if isinstance(dict_to_sort, dict):
        dict_to_sort = {key: dict_to_sort[key] for key in sorted(dict_to_sort.keys())}
        for key, value in dict_to_sort.items():
            dict_to_sort[key] = sort_dict(value)

    return dict_to_sort



"""
When run from either the command line or in library mode (note **not** as an
ESM Plugin), ``esm_archiving`` can be configured to how it looks for specific
files. The configuration file is called ``esm_archiving_config``, should be
written in YAML, and have the following format::

    echam:  # The model name
        archive: # archive seperator **required**
            # Frequency specification (how often
            # a datestamp is generated to look for)
            frequency: "1M"
            # Date format specification
            date_format: "%Y%m"


By default, ``esm_archive`` looks in the following locations:

    1. Current working directory
    2. Any files in the XDG Standard:
        https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html


If nothing is found, the program reverts to the hard-coded defaults, found in
``esm_archiving/esm_archiving/config.py``

.. note::

    In future, it might be changed that the program will look for an experiment
    specific configuration based upon the path it is given during the
    ``create`` or ``upload`` step.

Generating a configuration
~~~~~~~~~~~~~~~~~~~~~~~~~~

You can use the command line switches ``--write_local_config`` and
``--write_config`` to generate configuration files either in the current
working directory, or in the global directory for your user account defined by
the XDG standard (typically ~/.config/esm_archiving)::

    $ esm_archive --write_local_config
    Writing local (experiment) configuration...

    $ esm_archive --write_config
    Writing global (user) configuration...

"""
import logging
import os

from xdgenvpy import XDGPedanticPackage
import yaml

# Add XDG Standard
xdg = XDGPedanticPackage("esm_archiving")
config_dirs = xdg.XDG_CONFIG_DIRS.split(":")
config_dirs = [l + "/esm_archiving" for l in config_dirs]
config_dirs.insert(0, os.curdir)

# Defaults:
CONFIG_FNAME = "esm_archiving_config"
DEFAULT_CONFIG = {
    "echam": {"archive": {"frequency": "1M", "date_format": "%Y%m"}},
    "jsbach": {"archive": {"frequency": "1M", "date_format": "%Y%m"}},
    "hdmodel": {"archive": {"frequency": "1M", "date_format": "%Y%m"}},
    "fesom": {"archive": {"frequency": "1Y", "date_format": "%Y%m"}},
    "oasis3mct": {"archive": {"frequency": "1M", "date_format": "%Y%m"}},
    "general": {"database_file": str(xdg.XDG_DATA_HOME) + "/esm_archiving.db"},
}


def load_config():
    """
    Loads the configuration from one of the default configuration directories.
    If none can be found, returns the hard-coded default configuration.

    Returns
    -------
    dict
        A representation of the configuration used for archiving.
    """
    for loc in config_dirs:
        read_config_fname = CONFIG_FNAME
        try:
            with open(os.path.join(loc, read_config_fname)) as source:
                config = yaml.load(source, Loader=yaml.FullLoader)
                return config
        except IOError:
            pass
    return DEFAULT_CONFIG


def write_config_yaml(path=None):
    if not path:
        path = xdg.XDG_CONFIG_DIRS.split(":")[0] + "/esm_archiving"
    logging.debug("Opening %s for writing..." % os.path.join(path, CONFIG_FNAME))
    with open(os.path.join(path, CONFIG_FNAME), "w") as config_file:
        logging.debug("...dumping...")
        yaml.dump(DEFAULT_CONFIG, config_file)
        logging.debug("...done!")



#!/usr/bin/env python
"""
A small wrapper that combines the shell interface and the Python interface
"""

# Import from Python Standard Library
import argparse
from loguru import logger

from .initialization import *
from .tests import *
from .test_utilities import *
from .read_shipped_data import *

import os
import sys


def main():
    # Logger
    logger.remove()
    logger.add(
        sys.stderr,
        filter={"": "WARNING", "esm_tests": "DEBUG"},
        format="<level>{message}</level>",
    )
    if os.environ.get("CI", False):
        logger.add(
            "out.log",
            filter={"": "WARNING", "esm_tests": "DEBUG"},
            backtrace=True,
            diagnose=True,
        )

    # Parsing
    parser = argparse.ArgumentParser(description="Automatic testing for ESM-Tools devs")
    parser.add_argument(
        "-n",
        "--no-user",
        default=False,
        help="Avoid loading user config",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--check",
        default=False,
        help="Check mode on (does not compile or run, but produces some files that can "
        + "be compared to previous existing files in 'last_tested' folder)",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--delete",
        default=False,
        help="Delete previous tests",
        action="store_true",
    )
    # parser.add_argument(
    #    "-k",
    #    "--keep",
    #    default=False,
    #    help="Keep run_, outdata and restart folders for runs",
    #    action="store_true",
    # )
    parser.add_argument(
        "-s",
        "--save",
        default="Not defined",
        help="Save files for comparisson in 'last_tested' folder",
    )
    parser.add_argument(
        "-t",
        "--state",
        default=False,
        help="Print the state stored in state.yaml",
        action="store_true",
    )
    parser.add_argument(
        "-o",
        "--hold",
        default=False,
        help="Hold before operation, to give time to check the output",
        action="store_true",
    )
    parser.add_argument(
        "-b",
        "--bulletpoints",
        default=False,
        help="bullet points for printing the results",
        action="store_true",
    )

    args = vars(parser.parse_args())

    save_flag = args["save"]
    print_state = args["state"]
    delete_tests = args["delete"]

    info = {}

    info["ignore_user_info"] = args["no_user"]
    info["actually_compile"] = not args["check"]
    info["actually_run"] = not args["check"]
    # info["keep_run_folders"] = args["keep"]
    info["hold"] = args["hold"]
    info["bulletpoints"] = args["bulletpoints"]

    info["script_dir"] = os.path.join(os.path.dirname(os.path.realpath(__file__)), ".")
    info["last_tested_dir"] = get_last_tested_dir()
    info["this_computer"] = (
        determine_computer_from_hostname().split("/")[-1].replace(".yaml", "")
    )

    # Predefined for later
    user_scripts = dict(comp={}, run={})

    # Print state if necessary
    if print_state:
        current_state = get_state_yaml()
        print_results(current_state, info)
        sys.exit(1)

    # Get user info for testing
    info = user_config(info)

    # Define lines to be ignored during comparison
    try:
        info["ignore"] = get_ignore_compare_yaml()
    except FileNotFoundError as e:
        print("Whoops, that did not work... I was looking here:")
        print(f"{info['script_dir']}/ignore_compare.yaml")
        for f in os.listdir(info["script_dir"]):
            print(f)
        print(e)
        raise

    logger.debug(f"User info: {info.get('user')}")
    logger.debug(f"Actually compile: {info.get('actually_compile')}")
    logger.debug(f"Actually run: {info.get('actually_run')}")

    # Gather scripts
    info = get_scripts(info)

    # Complete scripts_info
    info = read_info_from_rs(info)

    # Delete previous test
    if delete_tests:
        del_prev_tests(info)

    # Compile
    comp_test(info)

    # Run
    run_test(info)

    # Print results
    print_results(format_results(info), info)

    # Save files
    if save_flag == "Not defined":
        save_files(info, False)
    elif save_flag == "true" or save_flag == "True":
        save_files(info, True)

    # yprint(info["scripts"])



import sys, copy, os
import esm_parser


def skip_cluster(cluster, config):
    gw_config = config["general"]["workflow"]
    clusterconf = gw_config["subjob_clusters"][cluster]

    """
    print(f"run_only {clusterconf.get('run_only', 'Error') }")
    print(f"skip_chunk_number {clusterconf.get('skip_chunk_number', -999)}")
    print(f"skip_run_number {clusterconf.get('skip_run_number', -999)}")
    print(f"chunk_number {config['general'].get('chunk_number', -998)}")
    print(f"run_number {config['general'].get('run_number', -998)}")
    print(f"last_run_in_chunk {config['general']['last_run_in_chunk']}")
    print(f"first_run_in_chunk {config['general']['first_run_in_chunk']}")
    """

    if clusterconf.get("run_only", "Error") == "last_run_in_chunk" and not config[
        "general"
    ].get("last_run_in_chunk", False):
        return True
    if clusterconf.get("run_only", "Error") == "first_run_in_chunk" and not config[
        "general"
    ].get("first_run_in_chunk", False):
        return True
    if clusterconf.get("skip_chunk_number", -999) == config["general"].get(
        "chunk_number", -998
    ):
        return True
    if clusterconf.get("skip_run_number", -999) == config["general"].get(
        "run_number", -998
    ):
        return True

    return False


def assemble_workflow(config):
    #
    config = init_total_workflow(config)
    config = collect_all_workflow_information(config)
    config = complete_clusters(config)
    config = order_clusters(config)
    config = prepend_newrun_job(config)

    if config["general"]["jobtype"] == "unknown":
        config["general"]["command_line_config"]["jobtype"] = config["general"][
            "workflow"
        ]["first_task_in_queue"]
        config["general"]["jobtype"] = config["general"]["workflow"][
            "first_task_in_queue"
        ]

    return config


def display_nicely(config):
    esm_parser.pprint_config(config["general"]["workflow"])
    return config


def prepend_newrun_job(config):
    gw_config = config["general"]["workflow"]
    first_cluster_name = gw_config["first_task_in_queue"]
    first_cluster = gw_config["subjob_clusters"][first_cluster_name]

    if not first_cluster.get("batch_or_shell", "Error") == "SimulationSetup":

        last_cluster_name = gw_config["last_task_in_queue"]
        last_cluster = gw_config["subjob_clusters"][last_cluster_name]

        new_first_cluster_name = "newrun"
        new_first_cluster = {
            "newrun": {
                "called_from": last_cluster_name,
                "run_before": first_cluster_name,
                "next_submit": [first_cluster_name],
                "subjobs": ["newrun_general"],
                "batch_or_shell": "SimulationSetup",
            }
        }

        last_cluster["next_submit"].append("newrun")
        last_cluster["next_submit"].remove(first_cluster_name)

        first_cluster["called_from"] = "newrun"

        gw_config["first_task_in_queue"] = "newrun"

        new_subjob = {
            "newrun_general": {
                "nproc": 1,
                "called_from": last_cluster_name,
                "run_before": first_cluster_name,
                "next_submit": [first_cluster_name],
                "subjob_cluster": "newrun",
            }
        }

        gw_config["subjob_clusters"].update(new_first_cluster)
        gw_config["subjobs"].update(new_subjob)

    return config

    #


def order_clusters(config):
    gw_config = config["general"]["workflow"]

    for subjob_cluster in gw_config["subjob_clusters"]:
        if not "next_submit" in gw_config["subjob_clusters"][subjob_cluster]:
            gw_config["subjob_clusters"][subjob_cluster]["next_submit"] = []

    for subjob_cluster in gw_config["subjob_clusters"]:
        if not "run_after" in gw_config["subjob_clusters"][subjob_cluster]:
            if not ("run_before" in gw_config["subjob_clusters"][subjob_cluster]):

                print(f"Don't know when to execute cluster {subjob_cluster}.")
                print(gw_config)
                sys.exit(-1)

        if "run_after" in gw_config["subjob_clusters"][subjob_cluster]:
            if "run_before" in gw_config["subjob_clusters"][subjob_cluster]:
                print(
                    f"Specifying both run_after and run_before for cluster {subjob_cluster} may lead to problems."
                )
                print(f"Please choose.")
                sys.exit(-1)
            if (
                not gw_config["subjob_clusters"][subjob_cluster]["run_after"]
                in gw_config["subjob_clusters"]
            ):
                print(f"Unknown cluster {subjob_cluster['run_after']}.")
                sys.exit(-1)

            calling_cluster = gw_config["subjob_clusters"][subjob_cluster]["run_after"]

            if (
                not subjob_cluster
                in gw_config["subjob_clusters"][calling_cluster]["next_submit"]
            ):
                gw_config["subjob_clusters"][calling_cluster]["next_submit"].append(
                    subjob_cluster
                )
            gw_config["subjob_clusters"][subjob_cluster][
                "called_from"
            ] = calling_cluster

            if calling_cluster == gw_config["last_task_in_queue"]:
                gw_config["last_task_in_queue"] = subjob_cluster

        if "run_before" in gw_config["subjob_clusters"][subjob_cluster]:
            if (
                not gw_config["subjob_clusters"][subjob_cluster]["run_before"]
                in gw_config["subjob_clusters"]
            ):
                print(f"Unknown cluster {subjob_cluster['run_before']}.")
                sys.exit(-1)

            called_cluster = gw_config["subjob_clusters"][subjob_cluster]["run_before"]

            if (
                not called_cluster
                in gw_config["subjob_clusters"][subjob_cluster]["next_submit"]
            ):
                gw_config["subjob_clusters"][subjob_cluster]["next_submit"].append(
                    called_cluster
                )
            gw_config["subjob_clusters"][called_cluster]["called_from"] = subjob_cluster

            if called_cluster == gw_config["first_task_in_queue"]:
                gw_config["first_task_in_queue"] = subjob_cluster

    if "next_run_triggered_by" in gw_config:
        gw_config["last_task_in_queue"] = gw_config["next_run_triggered_by"]

    first_cluster_name = gw_config["first_task_in_queue"]
    first_cluster = gw_config["subjob_clusters"][first_cluster_name]
    last_cluster_name = gw_config["last_task_in_queue"]
    last_cluster = gw_config["subjob_clusters"][last_cluster_name]

    if not first_cluster_name in last_cluster.get("next_submit", ["Error"]):
        last_cluster["next_submit"].append(first_cluster_name)
    if not last_cluster_name in first_cluster.get("called_from", ["Error"]):
        first_cluster["called_from"] = last_cluster_name

    return config


def complete_clusters(config):
    gw_config = config["general"]["workflow"]

    # First, complete the matching subjobs <-> clusters

    for subjob in gw_config["subjobs"]:
        subjob_cluster = gw_config["subjobs"][subjob]["subjob_cluster"]
        if not subjob_cluster in gw_config["subjob_clusters"]:
            gw_config["subjob_clusters"][subjob_cluster] = {}

        if not "subjobs" in gw_config["subjob_clusters"][subjob_cluster]:
            gw_config["subjob_clusters"][subjob_cluster]["subjobs"] = []

        gw_config["subjob_clusters"][subjob_cluster]["subjobs"].append(subjob)

    # Then, complete the resource information per cluster
    # determine whether a cluster is to be submitted to a batch system

    for subjob_cluster in gw_config["subjob_clusters"]:
        nproc_sum = nproc_max = 0
        clusterconf = gw_config["subjob_clusters"][subjob_cluster]
        for subjob in clusterconf["subjobs"]:
            subjobconf = gw_config["subjobs"][subjob]

            clusterconf = merge_single_entry_if_possible(
                "submit_to_batch_system", subjobconf, clusterconf
            )
            clusterconf = merge_single_entry_if_possible(
                "order_in_cluster", subjobconf, clusterconf
            )

            if subjobconf.get("submit_to_batch_system", False):
                clusterconf["batch_or_shell"] = "batch"
            elif subjobconf.get("script", False):
                clusterconf["batch_or_shell"] = "shell"

            clusterconf = merge_single_entry_if_possible(
                "run_on_queue", subjobconf, clusterconf
            )
            clusterconf = merge_single_entry_if_possible(
                "run_after", subjobconf, clusterconf
            )
            clusterconf = merge_single_entry_if_possible(
                "run_before", subjobconf, clusterconf
            )
            clusterconf = merge_single_entry_if_possible(
                "run_only", subjobconf, clusterconf
            )
            clusterconf = merge_single_entry_if_possible(
                "skip_run_number", subjobconf, clusterconf
            )
            clusterconf = merge_single_entry_if_possible(
                "skip_chunk_number", subjobconf, clusterconf
            )

            nproc_sum += subjobconf.get("nproc", 1)
            nproc_max = max(subjobconf.get("nproc", 1), nproc_max)

        if not "submit_to_batch_system" in clusterconf:
            clusterconf["submit_to_batch_system"] = False
        else:
            if not "run_on_queue" in clusterconf:
                print(
                    f"Information on target queue is missing in cluster {clusterconf}."
                )
                sys.exit(-1)

        if not clusterconf.get("batch_or_shell", False):
            clusterconf["batch_or_shell"] = "SimulationSetup"

        if not "order_in_cluster" in clusterconf:
            clusterconf["order_in_cluster"] = "sequential"

        if clusterconf["order_in_cluster"] == "concurrent":
            nproc = nproc_sum
        else:
            nproc = nproc_max
        clusterconf["nproc"] = nproc

    return config


def merge_single_entry_if_possible(entry, sourceconf, targetconf):
    if entry in sourceconf:
        if entry in targetconf and not sourceconf[entry] == targetconf[entry]:
            print(f"Mismatch found in {entry} for cluster {targetconf}")
            sys.exit(-1)
        targetconf[entry] = sourceconf[entry]
    return targetconf


def init_total_workflow(config):
    # add compute, tidy etc information already here!

    tasks = 0
    for model in config["general"]["valid_model_names"]:
        if "nproc" in config[model]:
            tasks += config[model]["nproc"]
        elif "nproca" in config[model] and "nprocb" in config[model]:
            tasks += config[model]["nproca"] * config[model]["nprocb"]
            if "nprocar" in config[model] and "nprocbr" in config[model]:
                if (
                    config[model]["nprocar"] != "remove_from_namelist"
                    and config[model]["nprocbr"] != "remove_from_namelist"
                ):
                    tasks += config[model]["nprocar"] * config[model]["nprocbr"]

    prepcompute = {
        "prepcompute": {
            "nproc": 1,
            "run_before": "compute",
        }
    }

    compute = {
        "compute": {
            "nproc": tasks,
            "run_before": "tidy",
            "submit_to_batch_system": config["general"].get(
                "submit_to_batch_system", True
            ),
            "run_on_queue": config["computer"]["partitions"]["compute"]["name"],
        }
    }

    # das ist nur vorübergehend
    tidy = {
        "tidy": {
            "nproc": 1,
            "run_after": "compute",
        }
    }

    if not "workflow" in config["general"]:
        config["general"]["workflow"] = {}
    if not "subjob_clusters" in config["general"]["workflow"]:
        config["general"]["workflow"]["subjob_clusters"] = {}
    if not "subjobs" in config["general"]["workflow"]:
        config["general"]["workflow"]["subjobs"] = prepcompute
        config["general"]["workflow"]["subjobs"].update(compute)
        config["general"]["workflow"]["subjobs"].update(tidy)

    if not "last_task_in_queue" in config["general"]["workflow"]:
        config["general"]["workflow"]["last_task_in_queue"] = "tidy"
    if not "first_task_in_queue" in config["general"]["workflow"]:
        config["general"]["workflow"]["first_task_in_queue"] = "prepcompute"

    if not "next_run_triggered_by" in config["general"]["workflow"]:
        config["general"]["workflow"]["next_run_triggered_by"] = "tidy"

    return config


def collect_all_workflow_information(config):

    for model in config:
        if "workflow" in config[model]:
            w_config = config[model]["workflow"]
            gw_config = config["general"]["workflow"]

            if "subjob_clusters" in w_config:
                for cluster in w_config["subjob_clusters"]:
                    if cluster in gw_config["subjob_clusters"]:
                        gw_config["subjob_clusters"][cluster] = merge_if_possible(
                            w_config["subjob_clusters"][cluster],
                            gw_config["subjob_clusters"][cluster],
                        )
                    else:
                        gw_config["subjob_clusters"][cluster] = copy.deepcopy(
                            w_config["subjob_clusters"][cluster],
                        )

            if "subjobs" in w_config:
                ref_config = copy.deepcopy(w_config)
                for subjob in list(copy.deepcopy(w_config["subjobs"])):

                    # subjobs (other than clusters) should be model specific
                    gw_config["subjobs"][subjob + "_" + model] = copy.deepcopy(
                        w_config["subjobs"][subjob]
                    )
                    if subjob in gw_config["subjobs"]:
                        del gw_config["subjobs"][subjob]
                    # make sure that the run_after and run_before refer to that cluster
                    for other_subjob in gw_config["subjobs"]:
                        if "run_after" in gw_config["subjobs"][other_subjob]:
                            if (
                                gw_config["subjobs"][other_subjob]["run_after"]
                                == subjob
                            ):
                                gw_config["subjobs"][other_subjob][
                                    "run_after"
                                ] == subjob + "_" + model
                        if "run_before" in gw_config["subjobs"][other_subjob]:
                            if (
                                gw_config["subjobs"][other_subjob]["run_before"]
                                == subjob
                            ):
                                gw_config["subjobs"][other_subjob][
                                    "run_before"
                                ] == subjob + "_" + model

                    # if not in another cluster, each subjob gets its own
                    if (
                        not "subjob_cluster"
                        in gw_config["subjobs"][subjob + "_" + model]
                    ):
                        gw_config["subjobs"][subjob + "_" + model][
                            "subjob_cluster"
                        ] = subjob  # + "_" + model

            if "next_run_triggered_by" in w_config:
                if not gw_config["next_run_triggered_by"] in [
                    "tidy",
                    w_config["next_run_triggered_by"],
                ]:
                    print(f"Mismatch found setting next_run_triggered_by for workflow.")
                    sys.exit(-1)
                else:
                    gw_config["next_run_triggered_by"] = w_config[
                        "next_run_triggered_by"
                    ]

    return config


def merge_if_possible(source, target):
    for entry in source:
        if entry in target:
            if not source[entry] == target[entry]:
                print(
                    f"Mismatch while trying to merge subjob_clusters {source} into {target}"
                )
                sys.exit(-1)
        else:
            target[entry] = source[entry]
    return target



"""
Documentation goes here
"""
import sys
import os

from . import config_initialization
from . import prepare
from . import prepexp
from . import workflow
from . import resubmit
from . import helpers
from . import logfiles
from . import prev_run

import esm_parser


class SimulationSetup(object):
    def __init__(self, command_line_config=None, user_config=None):

        if not command_line_config and not user_config:
            raise ValueError(
                "SimulationSetup needs to be initialized with either command_line_config or user_config."
            )

        user_config = config_initialization.init_first_user_config(
            command_line_config, user_config
        )

        self.config = config_initialization.complete_config_from_user_config(
            user_config
        )

        self.config = config_initialization.save_command_line_config(
            self.config, command_line_config
        )

        # self.config = workflow.assemble(self.config)

        self.config["prev_run"] = prev_run.PrevRunInfo(self.config)
        self.config = prepare.run_job(self.config)

        # esm_parser.pprint_config(self.config)
        # sys.exit(0)

    def __call__(self, kill_after_submit=True):
        # Trigger inspect functionalities
        if self.config["general"]["jobtype"] == "inspect":
            # esm_parser.pprint_config(self.config)
            self.inspect()
            helpers.end_it_all(self.config)

        # Run the preexp recipe
        self.config = prepexp.run_job(self.config)

        # self.pseudocall(kill_after_submit)
        # call to observe here..
        org_jobtype = str(self.config["general"]["jobtype"])
        self.config = logfiles.initialize_logfiles(self.config, org_jobtype)

        if self.config["general"]["submitted"]:
            old_stdout = sys.stdout
            old_stderr = sys.stderr
            sys.stdout = logfiles.logfile_handle
            sys.stderr = logfiles.logfile_handle

        if self.config["general"]["jobtype"] == "prepcompute":
            self.prepcompute()
        elif self.config["general"]["jobtype"] == "tidy":
            self.tidy()
        elif self.config["general"]["jobtype"] == "viz":
            self.viz()
        elif self.config["general"]["jobtype"].startswith("observe"):
            pid = self.config["general"]["command_line_config"].get(
                "launcher_pid", -666
            )
            if not pid == -666:
                self.observe()

            self.config["general"]["jobtype"] = self.config["general"][
                "jobtype"
            ].replace("observe_", "")
            # that last line is necessary so that maybe_resubmit knows which
            # cluster to look up in the workflow

        else:
            self.assembler()

        resubmit.maybe_resubmit(self.config)

        self.config = logfiles.finalize_logfiles(self.config, org_jobtype)

        if self.config["general"]["submitted"]:
            sys.stdout = old_stdout
            sys.stderr = old_stderr

        if kill_after_submit:
            if self.config["general"].get("experiment_over", False):
                helpers.write_to_log(self.config, ["# Experiment over"], message_sep="")
            helpers.end_it_all(self.config)

        return self.config["general"].get("experiment_over", False)

    #########################     OBSERVE      #############################################################

    def observe(self):

        from . import observe

        self.config = observe.run_job(self.config)

    def assembler(self):
        from . import assembler

        self.config = assembler.run_job(self.config)

    ###################################     TIDY      #############################################################
    def tidy(self):
        """
        Performs steps for tidying up a simulation after a job has finished and
        submission of following jobs.

        This method uses two lists, ``all_files_to_copy`` and
        ``all_listed_filetypes`` to sort finished data from the **current run
        folder** back to the **main experiment folder** and submit new
        **compute** and **post-process** jobs. Files for ``log``, ``mon``,
        ``outdata``, and ``restart_out`` are gathered. The program waits until
        the job completes or an error is found (See ~self.wait_and_observe).
        Then, if necessary, the coupler cleans up it's files (unless it's a
        standalone run), and the files in the lists are copied from the **work
        folder** to the **current run folder**. A check for unknown files is
        performed (see ~self.check_for_unknown_files), files are
        moved from the  the **current run folder** to the **main experiment
        folder**.
        """
        from . import tidy

        self.config = tidy.run_job(self.config)

    ###################################     INSPECT      #############################################################
    def inspect(self):
        from . import inspect

        print(f"Inspecting {self.config['general']['experiment_dir']}")
        self.config = inspect.run_job(self.config)

    ###################################     PREPCOMPUTE      #############################################################
    def prepcompute(self):
        """
        All steps needed for a model computation.

        Parameters
        ----------
        kill_after_submit : bool
            Default ``True``. If set, the entire Python instance is killed with
            a ``sys.exit()`` as the very last after job submission.
        """
        from . import prepcompute

        self.config = prepcompute.run_job(self.config)

    ###################################     VIZ     #############################################################

    def viz(self):
        """
        Starts the Viz job.

        Parameters
        ----------
        kill_after_submit: bool
            Default ``True``. If set, the entire Python instance is killed with ``sys.exit()``.
        """
        # NOTE(PG): Local import, not everyone will have viz yet...
        import esm_viz as viz

        self.config = viz.run_job(self.config)



import os
import sys
import subprocess
import shlex  # contains shlex.split that respects quoted strings

# deniz: it is better to use more pathlib in the future so that dir/path
# operations will be more portable (supported since Python 3.4, 2014)
import pathlib

from .software_package import software_package

import esm_environment
import esm_plugin_manager


# Yes, Type Hints. Python >= 3.5 supports them. Small steps towards stability,
# until Paul goes crazy and redoes everything in Go. Or Rust. Or Brainfuck
# (yes, that's not made up: https://en.wikipedia.org/wiki/Brainfuck)
# Docs for typing: https://docs.python.org/3/library/typing.html
def install(package: str) -> None:
    """
    Checks if a package is already installed in the system and if it's not, then it
    installs it.

    Parameters
    ----------
    package : str
        Name of the package or get operation. Can be a package name (e.g.
        ``numpy``) or a full pip address (e.g.
        ``git@https://github.com/esm-tools/esm_tools.git``)

    Returns
    -------
    None
    """
    package_name = package.split("/")[-1].replace(".git", "")
    installed_packages = esm_plugin_manager.find_installed_plugins()
    if not package_name in installed_packages:
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
        except OSError:  # PermissionDeniedError would be nicer...
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "--user", package]
            )


######################################################################################
################################# class "task" #######################################
######################################################################################


class Task:
    """What you can do with a software_package, e.g. comp-awicm-2.0"""

    def __init__(self, raw, setup_info, vcs, general, complete_config, parsed_args):
        if raw == "default":
            raw = ""
        if raw == "drytestall":
            # panic
            for package in setup_info.all_packages:
                for todo in package.targets:
                    try:
                        print(todo + "-" + package.raw_name)
                        newtask = Task(
                            todo + "-" + package.raw_name, setup_info, vcs, parsed_args
                        )
                        newtask.output_steps()
                    except:
                        print("Problem found with target " + newtask.raw_name)
                        sys.exit(1)
            sys.exit(0)

        self.required_plugins = {}
        for key in complete_config:
            if "required_plugins" in complete_config[key]:
                self.required_plugins[key] = complete_config[key]["required_plugins"]
        # Initialize the ``already_installed_plugins`` attribute if it does not exist
        general.already_installed_plugins = self.already_installed_plugins = getattr(
            general, "already_installed_plugins", []
        )

        if type(raw) == str:
            (
                self.todo,
                kind,
                model,
                version,
                self.only_subtask,
                self.raw_name,
            ) = setup_info.split_raw_target(raw, setup_info)
            self.package = software_package(
                (kind, model, version), setup_info, vcs, general
            )
        else:  # tupel:
            (self.todo, kind, model, version, self.only_subtask) = raw
            self.package = software_package(
                (kind, model, version), setup_info, vcs, general
            )
            self.raw_name = setup_info.assemble_raw_name(
                self.todo, kind, model, version
            )

        if kind == "components":
            self.env = esm_environment.esm_environment.EnvironmentInfos(
                "compiletime", complete_config, model
            )
        else:
            self.env = None
        if not self.todo in setup_info.meta_todos:
            self.check_if_target(setup_info)

        self.subtasks = self.get_subtasks(
            setup_info, vcs, general, complete_config, parsed_args
        )
        self.only_subtask = self.validate_only_subtask()
        self.ordered_tasks = self.order_subtasks(setup_info, vcs, general)

        self.will_download = self.check_if_download_task(setup_info)
        self.folders_after_download = self.download_folders()
        self.binaries_after_compile = self.compile_binaries()
        self.dir_list = self.list_required_dirs()
        self.command_list, self.shown_command_list = self.assemble_command_list()

        if parsed_args.get("verbose", False):
            self.output()

    def get_subtasks(self, setup_info, vcs, general, complete_config, parsed_args):
        subtasks = []
        if self.todo in setup_info.meta_todos:
            todos = setup_info.meta_command_order[self.todo]
        else:
            todos = [self.todo]
        for todo in todos:
            for subpackage in self.package.subpackages:
                if todo in subpackage.targets:
                    subtasks.append(
                        Task(
                            (
                                todo,
                                subpackage.kind,
                                subpackage.model,
                                subpackage.version,
                                None,
                            ),
                            setup_info,
                            vcs,
                            general,
                            complete_config,
                            parsed_args,
                        )
                    )
        if subtasks == [] and self.todo in setup_info.meta_todos:
            # if self.todo in setup_info.meta_todos:
            for todo in todos:
                if todo in self.package.targets:
                    subtasks.append(
                        Task(
                            (
                                todo,
                                self.package.kind,
                                self.package.model,
                                self.package.version,
                                None,
                            ),
                            setup_info,
                            vcs,
                            general,
                            complete_config,
                            parsed_args,
                        )
                    )
        return subtasks

    def validate_only_subtask(self):
        only = None
        if self.only_subtask:
            only = []
            for task in self.subtasks:
                if task.package.raw_name.startswith(self.only_subtask):
                    only.append(task)
            if self.package.raw_name.startswith(self.only_subtask):
                self.subtasks = []
                return None
            if only == []:
                print()
                print(
                    "Given subtask "
                    + self.only_subtask
                    + " is not a valid subtask of package "
                    + self.raw_name
                    + "."
                )
                print()
                sys.exit(0)

        return only

    def order_subtasks(self, setup_info, vcs, general):
        subtasks = self.subtasks
        if self.only_subtask:
            if self.only_subtask == "NONE":
                return []
            elif type(self.only_subtask) == str:
                return [self.only_subtask]
            else:
                subtasks = self.only_subtask
        if subtasks == []:
            return [self]
        if self.todo in setup_info.meta_todos:
            todos = setup_info.meta_command_order[self.todo]
        else:
            todos = [self.todo]

        ordered_tasks = []
        for todo in todos:
            for task in subtasks:
                if task.todo == todo and task.package.bin_type == "lib":
                    ordered_tasks.append(task)
            for task in subtasks:
                if task.todo == todo and not task.package.bin_type == "lib":
                    ordered_tasks.append(task)  #
        if self.package.kind == "components" and not self.only_subtask:
            ordered_tasks.append(self)
        return ordered_tasks

    def check_if_download_task(self, setup_info):
        if self.todo == "get":
            return True
        if self.todo in setup_info.meta_todos:
            if "get" in setup_info.meta_command_order[self.todo]:
                return True
        return False

    def download_folders(self):
        # if self.package.kind in ["setups", "couplings"]:
        if self.package.subpackages:
            dir_list = [self.package.raw_name]
            for task in self.ordered_tasks:
                if (
                    self.package.raw_name + "/" + task.package.destination
                    not in dir_list
                ):
                    dir_list.append(
                        self.package.raw_name + "/" + task.package.destination
                    )
        else:
            dir_list = []
            for task in self.ordered_tasks:
                if task.package.destination not in dir_list:
                    dir_list.append(task.package.destination)
        return dir_list

    def compile_binaries(self):
        file_list = []
        for task in self.ordered_tasks:
            for binfile in task.package.bin_names:
                if (
                    self.package.raw_name
                    + "/"
                    + task.package.bin_type
                    + "/"
                    + binfile.split("/")[-1]
                    not in file_list
                ):
                    file_list.append(
                        self.package.raw_name
                        + "/"
                        + task.package.bin_type
                        + "/"
                        + binfile.split("/")[-1]
                    )
        return file_list

    def list_required_dirs(self):
        toplevel = self.package.raw_name
        if self.package.kind in ["setups", "couplings"] and self.will_download:
            dir_list = [self.package.raw_name]
        else:
            dir_list = []
        for task in self.ordered_tasks:
            if task.todo == "comp":
                if task.package.bin_names:
                    newdir = toplevel + "/" + task.package.bin_type
                    if newdir not in dir_list:
                        dir_list.append(newdir)
        return dir_list

    def assemble_command_list(self):
        command_list = []
        toplevel = self.package.destination
        # if self.package.kind in ["setups", "couplings"]:
        if self.package.subpackages:  # ???
            command_list.append("mkdir -p " + toplevel)
            command_list.append("cd " + toplevel)
            toplevel = "."
        real_command_list = command_list.copy()
        for task in self.ordered_tasks:
            if task.todo in ["get"]:
                if task.package.command_list[task.todo] is not None:
                    for command in task.package.command_list[task.todo]:
                        command_list.append(command)
                        real_command_list.append(command)

        if self.package.coupling_changes:
            for change in self.package.coupling_changes:
                command_list.append(change)
                real_command_list.append(change)

        for task in self.ordered_tasks:
            if task.todo not in ["get"]:
                if task.todo in ["conf", "comp"]:
                    # if self.package.kind in ["setups", "couplings"]:
                    if task.package.kind not in ["setups", "couplings"]:
                        if self.package.subpackages:
                            real_command_list.append(
                                "cp ../" + task.raw_name + "_script.sh ."
                            )
                        real_command_list.append("./" + task.raw_name + "_script.sh")
                else:
                    if task.package.command_list[task.todo] is not None:
                        for command in task.package.command_list[task.todo]:
                            real_command_list.append(command)
                if task.package.command_list[task.todo] is not None:
                    for command in task.package.command_list[task.todo]:
                        command_list.append(command)
                if task.todo == "comp":
                    if task.package.bin_names:
                        command_list.append(
                            "mkdir -p " + toplevel + "/" + task.package.bin_type
                        )
                        real_command_list.append(
                            "mkdir -p " + toplevel + "/" + task.package.bin_type
                        )
                        for binfile in task.package.bin_names:
                            # PG: Only copy if source and dest aren't the same!
                            # (Prevents cp: ‘/temp/test.txt’ and
                            # ‘/temp/test/test.txt’ are the same file)
                            toplevel_bin_path = (
                                toplevel
                                + "/"
                                + task.package.bin_type
                                + "/"
                                + binfile.split("/")[-1]
                            )
                            # MA: If there are already commands that will clean the
                            # bin folder, the following `if` is required to be true
                            clean_command_list = ["rm -f " + toplevel_bin_path]
                            clean_command = any(
                                cc in command_list for cc in clean_command_list
                            )

                            # deniz: bug fix for the fesom compilation issue
                            # at some point these string concats need to be
                            # replaced by pathlib
                            bin_path_target = pathlib.Path(
                                toplevel + "/" + task.package.bin_type
                            )
                            binary_file_path = pathlib.Path(
                                task.package.destination + "/" + binfile
                            )
                            binary_file_parent = binary_file_path.parent
                            # path2bin_origin = "/".join(binary_file_path.split('/')[:-1])

                            # deniz: don't copy the files if the paths are same.
                            # I think pathlib.Path.resolve() is a better option
                            # than simple string comparison
                            should_copy_files = (
                                binary_file_parent.resolve()
                                != bin_path_target.resolve()
                            )

                            # add the remaining conditions
                            should_copy_files = should_copy_files and (
                                not os.path.exists(toplevel_bin_path) or clean_command
                            )

                            if should_copy_files:
                                # deniz: I think " ".join() is a more Pythonic
                                # way to construct the command strings
                                cmd_str = "  ".join(
                                    ["cp", str(binary_file_path), str(bin_path_target)]
                                )
                                command_list.append(cmd_str)
                                real_command_list.append(cmd_str)

                elif task.todo == "clean":
                    if task.package.bin_names:
                        for binfile in task.package.bin_names:
                            command_list.append(
                                f"rm -f {toplevel}/{task.package.bin_type}/{binfile.split('/', -1)[-1]}"
                            )
                            real_command_list.append(
                                f"rm -f {toplevel}/{task.package.bin_type}/{binfile.split('/', -1)[-1]}"
                            )
        if task.todo in ["comp"]:
            for component in self.required_plugins:
                for plugin in self.required_plugins[component]:
                    # Install the plugin if is not already installed
                    if plugin not in self.already_installed_plugins:
                        # Actually only works because Paul put the gfw_creator
                        # required plugin for awiesm onto PyPI...
                        install(plugin)
                        self.already_installed_plugins.append(plugin)

        if self.package.kind in ["setups", "couplings"]:
            command_list.append("cd ..")
            real_command_list.append("cd ..")

        return real_command_list, command_list

    def cleanup_script(self):
        try:
            os.remove("./dummy_script.sh")
        except OSError:
            print("No dummy script to remove!")
        for task in self.ordered_tasks:
            if task.todo in ["conf", "comp"]:
                try:
                    os.remove("./" + task.raw_name + "_script.sh")
                except OSError:
                    print("No file to remove for ", task.raw_name)

    def check_if_target(self, setup_info):
        if not setup_info.has_target2(self.package, self.todo):
            setup_info.output_available_targets(self.raw_name)
            sys.exit(0)

    def check_requirements(self):
        if self.will_download:
            return True
        requirements = self.folders_after_download
        for folder in requirements:
            if not os.path.isdir(folder):
                print()
                print(
                    "Missing folder "
                    + folder
                    + " detected. Please run 'make get-"
                    + self.package.raw_name
                    + "' first."
                )
                print()
                sys.exit(0)
        return True

    def validate(self):
        self.check_requirements()

    def execute(self, ignore_errors=False):
        for task in self.ordered_tasks:
            if task.todo in ["conf", "comp"]:
                if task.package.kind == "components":
                    task.env.write_dummy_script()
                    newfile = task.env.add_commands(
                        task.package.command_list[task.todo], task.raw_name
                    )
                    if os.path.isfile(newfile):
                        os.chmod(newfile, 0o755)
        for command in self.command_list:
            if command.startswith("mkdir"):
                # os.system(command)
                subprocess.run(command.split(), check=not ignore_errors)
            elif command.startswith("cp "):
                subprocess.run(command.split(), check=not ignore_errors)
            elif command.startswith("cd ") and ";" not in command:
                os.chdir(command.replace("cd ", ""))
            # deniz: add pipe support
            elif "|" in command:
                # if there is a pipe in the command, then separate these in to
                # two parts. Eg. curl foo.tar.gz | tar zx
                curl_command, pipe_command = command.split("|")
                curl_process = subprocess.Popen(
                    curl_command.split(), stdout=subprocess.PIPE
                )
                output = subprocess.check_output(
                    pipe_command.split(), stdin=curl_process.stdout
                )
                curl_process.wait()
            else:
                # os.system(command)
                # deniz: I personally did not like the iterator and the list
                # having the same name. for com in command.split(';') would be
                # better IMHO
                for command in command.split(";"):
                    # seb-wahl: use shlex split as sed commands that use spaces
                    # need to be quoted, shlex split doesn't split quoted
                    # strings on spaces
                    # example: sed -i '/COUPLENEMOFOCI = /s/.FALSE./.TRUE./g' oifs-43r3-foci/src/ifs/module/yommcc.F90
                    # will fail if the "'" is removed
                    command_spl = shlex.split(command)
                    if "cd" == command_spl[0]:
                        os.chdir(command_spl[1])
                    else:
                        subprocess.run(
                            command_spl,
                            check=True,
                            shell=(
                                command.startswith("./") and command.endswith(".sh")
                            ),
                        )

    def output(self):
        print()
        subtasklist = []
        osubtasklist = []
        self.package.output()
        print("    Todo: ", self.todo)
        if self.only_subtask:
            if self.only_subtask == "NONE":
                print("    NO VALID SUBTASKS!!!")
            else:
                for subtask in self.only_subtask:
                    print("    Only Subtask:", subtask.package.raw_name)
        for subtask in self.subtasks:
            subtasklist.append(subtask.raw_name)
        for osubtask in self.ordered_tasks:
            osubtasklist.append(osubtask.raw_name)
        if not subtasklist == []:
            print("    Subtasks:", subtasklist)
        if not osubtasklist == []:
            print("    Ordered Subtasks:", osubtasklist)
        self.output_steps()
        if not self.folders_after_download == []:
            print("    The following folders should exist after download:")
            for folder in self.folders_after_download:
                print("        ", folder)
        if not self.binaries_after_compile == []:
            print("    The following files should exist after compiling:")
            for binfile in self.binaries_after_compile:
                print("        ", binfile)

    def output_steps(self):
        if not self.command_list == []:
            print("    Executing commands in this order:")
            for command in self.shown_command_list:
                print("        ", command)



import filecmp
import os
import glob
import sys

from esm_parser import pprint_config
from .helpers import evaluate
from .prepcompute import _show_simulation_info
from .namelists import Namelist
from . import workflow


def run_job(config):
    config = evaluate(config, "inspect", "inspect_recipe")
    return config


def inspect_workflow(config):
    if config["general"]["inspect"] == "workflow":

        config = workflow.display_nicely(config)
        sys.exit(0)
    return config


def inspect_overview(config):
    if config["general"]["inspect"] == "overview":
        config = _show_simulation_info(config)
        sys.exit(0)
    return config


def inspect_namelists(config):
    if config["general"]["inspect"] == "namelists":
        for model in config["general"]["valid_model_names"]:
            config[model] = Namelist.nmls_load(config[model])
            config[model] = Namelist.nmls_output(config[model])
        sys.exit(0)
    return config


def inspect_config(config):
    if config["general"]["inspect"] == "config":
        pprint_config(config)
        sys.exit(0)
    return config


def inspect_size(config):
    if config["general"]["inspect"] == "size":
        total_size = dir_size(config["general"]["experiment_dir"])
        unit = "B"
        if total_size >= 1024:
            total_size = total_size / 1024.0
            unit = "kB"
        if total_size >= 1024:
            total_size = total_size / 1024.0
            unit = "MB"
        if total_size >= 1024:
            total_size = total_size / 1024.0
            unit = "GB"
        if total_size >= 1024:
            total_size = total_size / 1024.0
            unit = "TB"
        print(f"Total size: {total_size:.2f} {unit}")
        sys.exit(0)
    return config


def inspect_folder(config):
    checkpath = config["general"]["thisrun_dir"] + "/" + config["general"]["inspect"]
    if os.path.isdir(checkpath):
        all_files = os.listdir(checkpath)
        print(f"Files in folder {checkpath}:")
        for thisfile in sorted(all_files):
            print(f" -- {thisfile}")
        sys.exit(0)
    return config


def inspect_file(config):
    exclude = ["work"]
    knownfiles = {}
    search_dir = config["general"]["thisrun_dir"]
    if config["general"]["inspect"] == "lastlog":
        maybe_file = config["computer"]["thisrun_logfile"].replace(
            "@jobtype@", "compute"
        )
    elif config["general"]["inspect"] == "explog":
        maybe_file = (
            f"{config['general']['expid']}_{config['general']['setup_name']}.log"
        )
        search_dir = config["general"]["experiment_dir"]
    elif config["general"]["inspect"] == "datefile":
        maybe_file = (
            f"{config['general']['expid']}_{config['general']['setup_name']}.date"
        )
        search_dir = config["general"]["experiment_dir"]
    else:
        maybe_file = config["general"]["inspect"]

    for path, subdirs, files in os.walk(search_dir):
        if not path.endswith("work"):  # skip work for now
            for full_filepath in glob.iglob(os.path.join(path, maybe_file)):
                cat_file(full_filepath)
                knownfiles.update({os.path.basename(full_filepath): full_filepath})

    for path, subdirs, files in os.walk(config["general"]["thisrun_dir"] + "/work"):
        for full_filepath in glob.iglob(os.path.join(path, maybe_file)):
            somefile = os.path.basename(full_filepath)
            if somefile in knownfiles:
                if filecmp.cmp(knownfiles[somefile], full_filepath):
                    print(
                        f"File {full_filepath} is identical to {knownfiles[somefile]}, skipping."
                    )
                    continue
                else:
                    print(f"File {full_filepath} differs from {knownfiles[somefile]}.")
            cat_file(full_filepath)
    return config


def dir_size(somepath):
    size = 0
    for path, subdirs, files in os.walk(somepath):
        for somefile in files:
            full_filepath = os.path.join(path, somefile)
            if not os.path.islink(full_filepath):
                size += os.path.getsize(full_filepath)
    return size


def cat_file(full_filepath):
    if os.path.isfile(full_filepath):
        print(f"Content of {full_filepath}:")
        with open(full_filepath, "r") as log:
            print(log.read())



"""
Functionality for displaying the version number
"""
import sys
import click


@click.version_option()
@click.group()
def main(args=None):
    """Console script for esm_tools"""
    return 0


if __name__ == "__main__":
    sys.exit(main())



from sqlalchemy import create_engine, Column, Integer, String, Sequence, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import QueuePool

import os

database_file = os.path.expanduser("~") + "/.esm_tools/esm_master.db"
if not os.path.isdir(os.path.expanduser("~") + "/.esm_tools"):
    os.mkdir(os.path.expanduser("~") + "/.esm_tools")

from esm_database import location_database

# engine = create_engine('sqlite:///' + database_file, echo = False, poolclass = NullPool)
engine = create_engine("sqlite:///" + database_file, echo=False, poolclass=QueuePool)
base = declarative_base()


class installation(base):
    __tablename__ = "installation"

    id = Column(Integer, Sequence("user_id_seq"), primary_key=True)
    timestamp = Column(DateTime)

    setup_name = Column(String)
    action = Column(String, default="download")
    folder = Column(String, default="none yet")

    location_database.register("installation", database_file, "esm_master")

    @staticmethod
    def topline():
        print()
        print("INSTALLATIONS:")
        print(
            "{0: >4}".format("ID")
            + "   "
            + "{0: >17}".format("timestamp")
            + "   "
            + "{0: >10}".format("setup")
            + "   "
            + "{0: >10}".format("action")
            + "   "
            + "{0: >45}".format("folder")
            + "   "
        )

    @staticmethod
    def nicer_output(run):
        print("ID: " + str(run.id) + ":")
        print("     Timestamp: " + run.timestamp.strftime("%x %X"))
        print("     Setup: " + run.setup_name)
        print("     Action performed: " + run.action)
        print("     Installation in folder: " + run.folder)

    def __repr__(self):
        setup = (self.setup_name[:10]) if len(self.setup_name) > 10 else self.setup_name
        return "%4s   %17s   %10s   %10s   %45s" % (
            str(self.id),
            str(self.timestamp.strftime("%x-%X")),
            setup,
            self.action,
            self.folder,
        )


base.metadata.create_all(engine)
connection = engine.connect()
Session = sessionmaker(bind=engine)
Session.configure(bind=engine)
session = Session()
